# -*- coding: utf-8; -*-

import datetime
from unittest.mock import patch, MagicMock

from wuttjamaican.reports import Report

import colander
from pyramid.httpexceptions import HTTPNotFound

from wuttaweb.views import reports as mod
from wuttaweb.testing import WebTestCase


class SomeRandomReport(Report):
    """
    This report shows something random.
    """
    report_key = 'testing_some_random'
    report_title = "Random Test Report"

    def add_params(self, schema):

        schema.add(colander.SchemaNode(
            colander.String(),
            name='foo',
            missing=colander.null))

        schema.add(colander.SchemaNode(
            colander.Date(),
            name='start_date',
            missing=colander.null))

    def get_output_columns(self):
        return ['foo']

    def make_data(self, params, **kwargs):
        return {
            'output_title': "Testing Output",
            'data': [{'foo': 'bar'}],
        }


class TestReportViews(WebTestCase):

    def make_view(self):
        return mod.ReportView(self.request)

    def test_includeme(self):
        self.pyramid_config.include('wuttaweb.views.reports')

    def test_get_grid_data(self):
        view = self.make_view()
        providers = dict(self.app.providers)
        providers['wuttatest'] = MagicMock(report_modules=['tests.views.test_reports'])
        with patch.object(self.app, 'providers', new=providers):

            data = view.get_grid_data()
            self.assertIsInstance(data, list)
            self.assertTrue(data) # 1+ reports

    def test_normalize_report(self):
        view = self.make_view()
        report = SomeRandomReport(self.config)
        normal = view.normalize_report(report)
        help_text = normal.pop('help_text').strip()
        self.assertEqual(help_text, "This report shows something random.")
        self.assertEqual(normal, {
            'report_key': 'testing_some_random',
            'report_title': "Random Test Report",
        })

    def test_configure_grid(self):
        view = self.make_view()
        grid = view.make_model_grid()
        self.assertIn('report_title', grid.searchable_columns)
        self.assertIn('help_text', grid.searchable_columns)

    def test_get_instance(self):
        view = self.make_view()
        providers = {
            'wuttatest': MagicMock(report_modules=['tests.views.test_reports']),
        }
        with patch.object(self.app, 'providers', new=providers):

            # normal
            with patch.object(self.request, 'matchdict', new={'report_key': 'testing_some_random'}):
                report = view.get_instance()
                self.assertIsInstance(report, dict)
                self.assertEqual(report['report_key'], 'testing_some_random')
                self.assertEqual(report['report_title'], "Random Test Report")

            # not found
            with patch.object(self.request, 'matchdict', new={'report_key': 'this-should_notEXIST'}):
                self.assertRaises(HTTPNotFound, view.get_instance)

    def test_get_instance_title(self):
        view = self.make_view()
        result = view.get_instance_title({'report_title': 'whatever'})
        self.assertEqual(result, 'whatever')

    def test_view(self):
        self.pyramid_config.add_route('home', '/')
        self.pyramid_config.add_route('login', '/auth/login')
        self.pyramid_config.add_route('reports', '/reports/')
        self.pyramid_config.add_route('reports.view', '/reports/{report_key}')
        view = self.make_view()
        providers = dict(self.app.providers)
        providers['wuttatest'] = MagicMock(report_modules=['tests.views.test_reports'])
        with patch.object(self.app, 'providers', new=providers):
            with patch.object(self.request, 'matchdict', new={'report_key': 'testing_some_random'}):

                # initial view
                response = view.view()
                self.assertEqual(response.status_code, 200)
                # nb. there's a button in there somewhere, but no output title
                self.assertIn("Run Report", response.text)
                self.assertNotIn("Testing Output", response.text)

                # run the report
                with patch.object(self.request, 'GET', new={
                        '__start__': 'start_date:mapping',
                        'date': '2025-01-11',
                        '__end__': 'start_date',
                }):
                    response = view.view()
                    self.assertEqual(response.status_code, 200)
                    # nb. there's a button in there somewhere, *and* an output title
                    self.assertIn("Run Report", response.text)
                    self.assertIn("Testing Output", response.text)

    def test_configure_form(self):
        view = self.make_view()
        providers = dict(self.app.providers)
        providers['wuttatest'] = MagicMock(report_modules=['tests.views.test_reports'])
        with patch.object(self.app, 'providers', new=providers):

            with patch.object(self.request, 'matchdict', new={'report_key': 'testing_some_random'}):
                report = view.get_instance()
                form = view.make_model_form(report)
                self.assertIn('help_text', form.readonly_fields)
                self.assertIn('foo', form)

    def test_normalize_columns(self):
        view = self.make_view()

        columns = view.normalize_columns(['foo'])
        self.assertEqual(columns, [
            {'name': 'foo', 'label': 'foo'},
        ])

        columns = view.normalize_columns([{'name': 'foo'}])
        self.assertEqual(columns, [
            {'name': 'foo', 'label': 'foo'},
        ])

        columns = view.normalize_columns([{'name': 'foo', 'label': "FOO"}])
        self.assertEqual(columns, [
            {'name': 'foo', 'label': 'FOO'},
        ])

        columns = view.normalize_columns([{'name': 'foo', 'label': "FOO", 'numeric': True}])
        self.assertEqual(columns, [
            {'name': 'foo', 'label': 'FOO', 'numeric': True},
        ])

    def test_run_report(self):
        view = self.make_view()
        providers = dict(self.app.providers)
        providers['wuttatest'] = MagicMock(report_modules=['tests.views.test_reports'])
        with patch.object(self.app, 'providers', new=providers):

            with patch.object(self.request, 'matchdict', new={'report_key': 'testing_some_random'}):
                report = view.report_handler.get_report('testing_some_random')
                normal = view.normalize_report(report)
                form = view.make_model_form(normal)

                # typical
                context = view.run_report(report, {'form': form})
                self.assertEqual(sorted(context['report_params']), ['foo', 'start_date'])
                self.assertEqual(context['report_data'], {
                    'output_title': "Testing Output",
                    'data': [{'foo': 'bar'}],
                })
                self.assertIn('report_generated', context)

                # invalid params
                with patch.object(self.request, 'GET', new={'start_date': 'NOT_GOOD'}):
                    context = view.run_report(report, {'form': form})
                    self.assertNotIn('report_params', context)
                    self.assertNotIn('report_data', context)
                    self.assertNotIn('report_generated', context)

                # custom formatter
                with patch.object(report, 'get_output_columns') as get_output_columns:
                    get_output_columns.return_value = [
                        'foo',
                        {'name': 'start_date',
                         'formatter': lambda val: "FORMATTED VALUE"},
                    ]

                    with patch.object(report, 'make_data') as make_data:
                        make_data.return_value = [
                            {'foo': 'bar', 'start_date': datetime.date(2025, 1, 11)},
                        ]

                        context = view.run_report(report, {'form': form})
                        get_output_columns.assert_called_once_with()
                        self.assertEqual(len(context['report_columns']), 2)
                        self.assertEqual(context['report_columns'][0]['name'], 'foo')
                        self.assertEqual(context['report_columns'][1]['name'], 'start_date')
                        self.assertEqual(context['report_data'], {
                            'output_title': "Random Test Report",
                            'data': [{'foo': 'bar', 'start_date': 'FORMATTED VALUE'}],
                        })

    def test_download_data(self):
        view = self.make_view()
        providers = dict(self.app.providers)
        providers['wuttatest'] = MagicMock(report_modules=['tests.views.test_reports'])
        with patch.object(self.app, 'providers', new=providers):
            with patch.object(self.request, 'matchdict', new={'report_key': 'testing_some_random'}):

                params, columns, data = view.get_download_data()
                self.assertEqual(params, {})
                self.assertEqual(columns, [{'name': 'foo', 'label': 'foo'}])
                self.assertEqual(data, {
                    'output_title': "Testing Output",
                    'data': [{'foo': 'bar'}],
                })

    def test_download_path(self):
        view = self.make_view()
        data = {'output_title': "My Report"}
        path = view.get_download_path(data, 'csv')
        self.assertTrue(path.endswith('My Report.csv'))



# -*- coding: utf-8; -*-

import datetime
from unittest.mock import patch, MagicMock

from sqlalchemy import orm
from pyramid.httpexceptions import HTTPFound

from wuttjamaican.db import model
from wuttjamaican.batch import BatchHandler
from wuttaweb.views import MasterView, batch as mod
from wuttaweb.progress import SessionProgress
from wuttaweb.testing import WebTestCase


class MockBatch(model.BatchMixin, model.Base):
    __tablename__ = 'testing_batch_mock'

class MockBatchRow(model.BatchRowMixin, model.Base):
    __tablename__ = 'testing_batch_mock_row'
    __batch_class__ = MockBatch

MockBatch.__row_class__ = MockBatchRow

class MockBatchHandler(BatchHandler):
    model_class = MockBatch


class TestBatchMasterView(WebTestCase):

    def setUp(self):
        self.setup_web()

        # nb. create MockBatch, MockBatchRow
        model.Base.metadata.create_all(bind=self.session.bind)

    def make_handler(self):
        return MockBatchHandler(self.config)

    def make_view(self):
        return mod.BatchMasterView(self.request)

    def test_get_batch_handler(self):
        self.assertRaises(NotImplementedError, mod.BatchMasterView, self.request)

        with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=42):
            view = mod.BatchMasterView(self.request)
            self.assertEqual(view.batch_handler, 42)

    def test_get_fallback_templates(self):
        handler = MockBatchHandler(self.config)
        with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=handler):
            view = self.make_view()
            templates = view.get_fallback_templates('view')
            self.assertEqual(templates, [
                '/batch/view.mako',
                '/master/view.mako',
            ])

    def test_render_to_response(self):
        model = self.app.model
        handler = MockBatchHandler(self.config)

        user = model.User(username='barney')
        self.session.add(user)
        batch = handler.make_batch(self.session, created_by=user)
        self.session.add(batch)
        self.session.flush()

        with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=handler):
            with patch.object(MasterView, 'render_to_response') as render_to_response:
                view = self.make_view()
                response = view.render_to_response('view', {'instance': batch})
                self.assertTrue(render_to_response.called)
                context = render_to_response.call_args[0][1]
                self.assertIs(context['batch'], batch)
                self.assertIs(context['batch_handler'], handler)

    def test_configure_grid(self):
        handler = MockBatchHandler(self.config)
        with patch.multiple(mod.BatchMasterView, create=True, model_class=MockBatch):
            with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=handler):
                view = mod.BatchMasterView(self.request)
                grid = view.make_model_grid()
                # nb. coverage only; tests nothing
                view.configure_grid(grid)

    def test_render_batch_id(self):
        handler = MockBatchHandler(self.config)
        with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=handler):
            view = mod.BatchMasterView(self.request)
            batch = MockBatch(id=42)

            result = view.render_batch_id(batch, 'id', 42)
            self.assertEqual(result, '00000042')

            result = view.render_batch_id(batch, 'id', None)
            self.assertIsNone(result)

    def test_get_instance_title(self):
        handler = MockBatchHandler(self.config)
        with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=handler):
            view = mod.BatchMasterView(self.request)

            batch = MockBatch(id=42)
            result = view.get_instance_title(batch)
            self.assertEqual(result, "00000042")

            batch = MockBatch(id=43, description="runnin some numbers")
            result = view.get_instance_title(batch)
            self.assertEqual(result, "00000043 runnin some numbers")

    def test_configure_form(self):
        handler = MockBatchHandler(self.config)
        with patch.multiple(mod.BatchMasterView, create=True, model_class=MockBatch):
            with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=handler):
                view = mod.BatchMasterView(self.request)

                # creating
                with patch.object(view, 'creating', new=True):
                    form = view.make_model_form(model_instance=None)
                    view.configure_form(form)

                batch = MockBatch(id=42)

                # viewing
                with patch.object(view, 'viewing', new=True):
                    form = view.make_model_form(model_instance=batch)
                    view.configure_form(form)

                # editing
                with patch.object(view, 'editing', new=True):
                    form = view.make_model_form(model_instance=batch)
                    view.configure_form(form)

                # deleting
                with patch.object(view, 'deleting', new=True):
                    form = view.make_model_form(model_instance=batch)
                    view.configure_form(form)

                # viewing (executed)
                batch.executed = datetime.datetime.now()
                with patch.object(view, 'viewing', new=True):
                    form = view.make_model_form(model_instance=batch)
                    view.configure_form(form)

    def test_objectify(self):
        handler = MockBatchHandler(self.config)
        with patch.multiple(mod.BatchMasterView, create=True, model_class=MockBatch):
            with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=handler):
                with patch.object(mod.BatchMasterView, 'Session', return_value=self.session):
                    view = mod.BatchMasterView(self.request)

                    # create batch
                    with patch.object(view, 'creating', new=True):
                        form = view.make_model_form(model_instance=None)
                        form.validated = {}
                        batch = view.objectify(form)
                        self.assertIsInstance(batch.id, int)
                        self.assertTrue(batch.id > 0)

                    # edit batch
                    with patch.object(view, 'editing', new=True):
                        with patch.object(view.batch_handler, 'make_batch') as make_batch:
                            form = view.make_model_form(model_instance=batch)
                            form.validated = {'description': 'foo'}
                            self.assertIsNone(batch.description)
                            batch = view.objectify(form)
                            self.assertEqual(batch.description, 'foo')

    def test_redirect_after_create(self):
        self.pyramid_config.add_route('mock_batches.view', '/batch/mock/{uuid}')
        handler = MockBatchHandler(self.config)
        with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=handler):
            with patch.multiple(mod.BatchMasterView, create=True,
                                model_class=MockBatch,
                                route_prefix='mock_batches'):
                view = mod.BatchMasterView(self.request)
                batch = MockBatch(id=42)

                # typically redirect to view batch
                result = view.redirect_after_create(batch)
                self.assertIsInstance(result, HTTPFound)

                # unless populating in which case thread is launched
                self.request.session.id = 'abcdefghijk'
                with patch.object(mod, 'threading') as threading:
                    thread = MagicMock()
                    threading.Thread.return_value = thread
                    with patch.object(view.batch_handler, 'should_populate', return_value=True):
                        with patch.object(view, 'render_progress') as render_progress:
                            view.redirect_after_create(batch)
                            self.assertTrue(threading.Thread.called)
                            thread.start.assert_called_once_with()
                            self.assertTrue(render_progress.called)

    def test_delete_instance(self):
        model = self.app.model
        handler = self.make_handler()

        user = model.User(username='barney')
        self.session.add(user)

        batch = handler.make_batch(self.session, created_by=user)
        self.session.add(batch)
        self.session.flush()

        with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=handler):
            view = self.make_view()

            self.assertEqual(self.session.query(MockBatch).count(), 1)
            view.delete_instance(batch)
            self.assertEqual(self.session.query(MockBatch).count(), 0)

    def test_populate_thread(self):
        model = self.app.model
        handler = MockBatchHandler(self.config)
        with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=handler):
            with patch.multiple(mod.BatchMasterView, create=True, model_class=MockBatch):
                view = mod.BatchMasterView(self.request)
                user = model.User(username='barney')
                self.session.add(user)
                batch = MockBatch(id=42, created_by=user)
                self.session.add(batch)
                self.session.commit()

                # nb. use our session within thread method
                with patch.object(self.app, 'make_session', return_value=self.session):

                    # nb. prevent closing our session
                    with patch.object(self.session, 'close') as close:

                        # without progress
                        view.populate_thread(batch.uuid)
                        close.assert_called_once_with()
                        close.reset_mock()

                        # with progress
                        self.request.session.id = 'abcdefghijk'
                        view.populate_thread(batch.uuid,
                                             progress=SessionProgress(self.request,
                                                                      'populate_mock_batch'))
                        close.assert_called_once_with()
                        close.reset_mock()

                        # failure to populate, without progress
                        with patch.object(view.batch_handler, 'do_populate', side_effect=RuntimeError):
                            view.populate_thread(batch.uuid)
                            close.assert_called_once_with()
                            close.reset_mock()

                        # failure to populate, with progress
                        with patch.object(view.batch_handler, 'do_populate', side_effect=RuntimeError):
                            view.populate_thread(batch.uuid,
                                                 progress=SessionProgress(self.request,
                                                                          'populate_mock_batch'))
                            close.assert_called_once_with()
                            close.reset_mock()

                        # failure for batch to appear
                        self.session.delete(batch)
                        self.session.commit()
                        # nb. should give up waiting after 1 second
                        self.assertRaises(RuntimeError, view.populate_thread, batch.uuid)

    def test_execute(self):
        self.pyramid_config.add_route('mock_batches.view', '/batch/mock/{uuid}')
        model = self.app.model
        handler = MockBatchHandler(self.config)

        user = model.User(username='barney')
        self.session.add(user)
        batch = handler.make_batch(self.session, created_by=user)
        self.session.add(batch)
        self.session.commit()

        with patch.multiple(mod.BatchMasterView, create=True,
                            model_class=MockBatch,
                            route_prefix='mock_batches',
                            get_batch_handler=MagicMock(return_value=handler),
                            get_instance=MagicMock(return_value=batch)):
            view = self.make_view()

            # batch executes okay
            response = view.execute()
            self.assertEqual(response.status_code, 302) # redirect to "view batch"
            self.assertFalse(self.request.session.peek_flash('error'))

            # but cannot be executed again
            response = view.execute()
            self.assertEqual(response.status_code, 302) # redirect to "view batch"
            # nb. flash has error this time
            self.assertTrue(self.request.session.peek_flash('error'))

    def test_get_row_model_class(self):
        handler = MockBatchHandler(self.config)
        with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=handler):
            view = self.make_view()

            self.assertRaises(AttributeError, view.get_row_model_class)

            # row class determined from batch class
            with patch.object(mod.BatchMasterView, 'model_class', new=MockBatch, create=True):
                cls = view.get_row_model_class()
                self.assertIs(cls, MockBatchRow)

            self.assertRaises(AttributeError, view.get_row_model_class)

            # view may specify row class
            with patch.object(mod.BatchMasterView, 'row_model_class', new=MockBatchRow, create=True):
                cls = view.get_row_model_class()
                self.assertIs(cls, MockBatchRow)

    def test_get_row_grid_data(self):
        handler = MockBatchHandler(self.config)
        model = self.app.model

        user = model.User(username='barney')
        self.session.add(user)

        batch = handler.make_batch(self.session, created_by=user)
        self.session.add(batch)
        row = handler.make_row()
        handler.add_row(batch, row)
        self.session.flush()

        with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=handler):

            view = self.make_view()
            self.assertRaises(AttributeError, view.get_row_grid_data, batch)

            Session = MagicMock(return_value=self.session)
            Session.query.side_effect = lambda m: self.session.query(m)
            with patch.multiple(mod.BatchMasterView, create=True,
                                Session=Session,
                                model_class=MockBatch):

                view = self.make_view()
                data = view.get_row_grid_data(batch)
                self.assertIsInstance(data, orm.Query)
                self.assertEqual(data.count(), 1)

    def test_configure_row_grid(self):
        handler = MockBatchHandler(self.config)
        model = self.app.model

        user = model.User(username='barney')
        self.session.add(user)

        batch = handler.make_batch(self.session, created_by=user)
        self.session.add(batch)
        row = handler.make_row()
        handler.add_row(batch, row)
        self.session.flush()

        with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=handler):

            Session = MagicMock(return_value=self.session)
            Session.query.side_effect = lambda m: self.session.query(m)
            with patch.multiple(mod.BatchMasterView, create=True,
                                Session=Session,
                                model_class=MockBatch):

                with patch.object(self.request, 'matchdict', new={'uuid': batch.uuid}):
                    view = self.make_view()
                    grid = view.make_row_model_grid(batch)
                    self.assertIn('sequence', grid.labels)
                    self.assertEqual(grid.labels['sequence'], "Seq.")

    def test_render_row_status(self):
        with patch.object(mod.BatchMasterView, 'get_batch_handler', return_value=None):
            view = self.make_view()
            row = MagicMock(foo=1, STATUS={1: 'bar'})
            self.assertEqual(view.render_row_status(row, 'foo', 1), 'bar')

    def test_defaults(self):
        # nb. coverage only
        with patch.object(mod.BatchMasterView, 'model_class', new=MockBatch, create=True):
            mod.BatchMasterView.defaults(self.pyramid_config)



# -*- coding: utf-8; -*-

from unittest.mock import patch

import colander
from pyramid.httpexceptions import HTTPNotFound

from wuttaweb.views import settings as mod
from wuttaweb.testing import WebTestCase


class TestAppInfoView(WebTestCase):

    def setUp(self):
        self.setup_web()
        self.pyramid_config.include('wuttaweb.views.essential')

    def make_view(self):
        return mod.AppInfoView(self.request)

    def test_get_grid_data(self):
        view = self.make_view()

        # empty data by default
        data = view.get_grid_data()
        self.assertEqual(data, [])

        # 'partial' request returns data
        self.request.GET = {'partial': '1'}
        data = view.get_grid_data()
        self.assertIsInstance(data, list)
        self.assertTrue(data)

    def test_index(self):
        # sanity/coverage check
        view = self.make_view()
        response = view.index()

    def test_configure_get_simple_settings(self):
        # sanity/coverage check
        view = self.make_view()
        simple = view.configure_get_simple_settings()

    def test_configure_get_context(self):
        # sanity/coverage check
        view = self.make_view()
        context = view.configure_get_context()


class TestSettingView(WebTestCase):

    def make_view(self):
        return mod.SettingView(self.request)

    def test_get_grid_data(self):

        # empty data by default
        view = self.make_view()
        query = view.get_grid_data(session=self.session)
        data = query.all()
        self.assertEqual(len(data), 0)

        # unless we save some settings
        self.app.save_setting(self.session, 'foo', 'bar')
        self.session.commit()
        query = view.get_grid_data(session=self.session)
        data = query.all()
        self.assertEqual(len(data), 1)

    def test_configure_grid(self):
        model = self.app.model
        view = self.make_view()
        grid = view.make_grid(model_class=model.Setting)
        self.assertFalse(grid.is_linked('name'))
        view.configure_grid(grid)
        self.assertTrue(grid.is_linked('name'))

    def test_configure_form(self):
        view = self.make_view()
        form = view.make_form(fields=view.get_form_fields())
        self.assertNotIn('value', form.required_fields)
        view.configure_form(form)
        self.assertIn('value', form.required_fields)
        self.assertFalse(form.required_fields['value'])

    def test_unique_name(self):
        model = self.app.model
        view = self.make_view()

        setting = model.Setting(name='foo')
        self.session.add(setting)
        self.session.commit()

        with patch.object(view, 'Session', return_value=self.session):

            # invalid if same name in data
            node = colander.SchemaNode(colander.String(), name='name')
            self.assertRaises(colander.Invalid, view.unique_name, node, 'foo')

            # but not if name belongs to current setting
            view.editing = True
            self.request.matchdict = {'name': 'foo'}
            node = colander.SchemaNode(colander.String(), name='name')
            self.assertIsNone(view.unique_name(node, 'foo'))



# -*- coding: utf-8; -*-

import datetime
from collections import OrderedDict
from enum import Enum
from unittest import TestCase
from unittest.mock import patch

import sqlalchemy as sa

from wuttjamaican.db.model import Base

from wuttaweb.grids import filters as mod
from wuttaweb.testing import WebTestCase


class TestGridFilter(WebTestCase):

    def setUp(self):
        self.setup_web()

        model = self.app.model
        self.sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'ggg'},
            {'name': 'foo4', 'value': 'ggg'},
            {'name': 'foo5', 'value': 'ggg'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in self.sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        self.sample_query = self.session.query(model.Setting)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop('factory', mod.GridFilter)
        kwargs['model_property'] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_constructor(self):
        model = self.app.model

        # verbs is not set by default, but can be set
        filtr = self.make_filter(model.Setting.name)
        self.assertFalse(hasattr(filtr, 'verbs'))
        filtr = self.make_filter(model.Setting.name, verbs=['foo', 'bar'])
        self.assertEqual(filtr.verbs, ['foo', 'bar'])

        # verb is not set by default, but can be set
        filtr = self.make_filter(model.Setting.name)
        self.assertFalse(hasattr(filtr, 'verb'))
        filtr = self.make_filter(model.Setting.name, verb='foo')
        self.assertEqual(filtr.verb, 'foo')

        # default verb is not set by default, but can be set
        filtr = self.make_filter(model.Setting.name)
        self.assertFalse(hasattr(filtr, 'default_verb'))
        filtr = self.make_filter(model.Setting.name, default_verb='foo')
        self.assertEqual(filtr.default_verb, 'foo')

    def test_repr(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.GridFilter)
        self.assertEqual(repr(filtr), "GridFilter(key='name', active=False, verb=None, value=None)")

    def test_get_verbs(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)
        self.assertFalse(hasattr(filtr, 'verbs'))
        self.assertEqual(filtr.default_verbs, ['equal', 'not_equal'])

        # by default, returns default verbs (plus 'is_any')
        self.assertEqual(filtr.get_verbs(), ['equal', 'not_equal', 'is_any'])

        # default verbs can be a callable
        filtr.default_verbs = lambda: ['foo', 'bar']
        self.assertEqual(filtr.get_verbs(), ['foo', 'bar', 'is_any'])

        # uses filtr.verbs if set
        filtr.verbs = ['is_true', 'is_false']
        self.assertEqual(filtr.get_verbs(), ['is_true', 'is_false', 'is_any'])

        # may add is/null verbs
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter,
                                 nullable=True)
        self.assertEqual(filtr.get_verbs(), ['equal', 'not_equal',
                                             'is_null', 'is_not_null',
                                             'is_any'])

        # filtr.verbs can be a callable
        filtr.nullable = False
        filtr.verbs = lambda: ['baz', 'blarg']
        self.assertEqual(filtr.get_verbs(), ['baz', 'blarg', 'is_any'])

    def test_get_default_verb(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)
        self.assertFalse(hasattr(filtr, 'verbs'))
        self.assertEqual(filtr.default_verbs, ['equal', 'not_equal'])
        self.assertEqual(filtr.get_verbs(), ['equal', 'not_equal', 'is_any'])

        # returns first verb by default
        self.assertEqual(filtr.get_default_verb(), 'equal')

        # returns filtr.verb if set
        filtr.verb = 'foo'
        self.assertEqual(filtr.get_default_verb(), 'foo')

        # returns filtr.default_verb if set
        # (nb. this overrides filtr.verb since the point of this
        # method is to return the *default* verb)
        filtr.default_verb = 'bar'
        self.assertEqual(filtr.get_default_verb(), 'bar')

    def test_get_verb_labels(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)
        self.assertFalse(hasattr(filtr, 'verbs'))
        self.assertEqual(filtr.get_verbs(), ['equal', 'not_equal', 'is_any'])

        labels = filtr.get_verb_labels()
        self.assertIsInstance(labels, dict)
        self.assertEqual(labels['equal'], "equal to")
        self.assertEqual(labels['not_equal'], "not equal to")
        self.assertEqual(labels['is_any'], "is any")

    def test_get_valueless_verbs(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)
        self.assertFalse(hasattr(filtr, 'verbs'))
        self.assertEqual(filtr.get_verbs(), ['equal', 'not_equal', 'is_any'])

        verbs = filtr.get_valueless_verbs()
        self.assertIsInstance(verbs, list)
        self.assertIn('is_any', verbs)

    def test_set_choices(self):
        model = self.app.model

        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)
        self.assertEqual(filtr.choices, {})
        self.assertEqual(filtr.data_type, 'string')

        class MockEnum(Enum):
            FOO = 'foo'
            BAR = 'bar'

        filtr.set_choices(MockEnum)
        self.assertEqual(filtr.choices, OrderedDict([
            ('FOO', 'foo'),
            ('BAR', 'bar'),
        ]))
        self.assertEqual(filtr.data_type, 'choice')

        filtr.set_choices(None)
        self.assertEqual(filtr.choices, {})
        self.assertEqual(filtr.data_type, 'string')

    def test_normalize_choices(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.name, factory=mod.AlchemyFilter)

        class MockEnum(Enum):
            FOO = 'foo'
            BAR = 'bar'

        choices = filtr.normalize_choices(MockEnum)
        self.assertEqual(choices, OrderedDict([
            ('FOO', 'foo'),
            ('BAR', 'bar'),
        ]))

        choices = filtr.normalize_choices(OrderedDict([
            ('first', '1'),
            ('second', '2'),
        ]))
        self.assertEqual(choices, OrderedDict([
            ('first', '1'),
            ('second', '2'),
        ]))

        choices = filtr.normalize_choices({
            'bbb': 'b',
            'aaa': 'a',
        })
        self.assertEqual(choices, OrderedDict([
            ('aaa', 'a'),
            ('bbb', 'b'),
        ]))

        choices = filtr.normalize_choices(['one', 'two', 'three'])
        self.assertEqual(choices, OrderedDict([
            ('one', 'one'),
            ('two', 'two'),
            ('three', 'three'),
        ]))

    def test_apply_filter(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value, factory=mod.StringAlchemyFilter)

        # default verb used as fallback
        # self.assertEqual(filtr.default_verb, 'contains')
        filtr.default_verb = 'contains'
        filtr.verb = None
        with patch.object(filtr, 'filter_contains', side_effect=lambda q, v: q) as filter_contains:
            filtered_query = filtr.apply_filter(self.sample_query, value='foo')
            filter_contains.assert_called_once_with(self.sample_query, 'foo')
        self.assertIsNone(filtr.verb)

        # filter verb used as fallback
        filtr.verb = 'equal'
        with patch.object(filtr, 'filter_equal', create=True, side_effect=lambda q, v: q) as filter_equal:
            filtered_query = filtr.apply_filter(self.sample_query, value='foo')
            filter_equal.assert_called_once_with(self.sample_query, 'foo')

        # filter value used as fallback
        filtr.verb = 'contains'
        filtr.value = 'blarg'
        with patch.object(filtr, 'filter_contains', side_effect=lambda q, v: q) as filter_contains:
            filtered_query = filtr.apply_filter(self.sample_query)
            filter_contains.assert_called_once_with(self.sample_query, 'blarg')

        # error if invalid verb
        self.assertRaises(mod.VerbNotSupported, filtr.apply_filter,
                          self.sample_query, verb='doesnotexist')
        filtr.verbs = ['doesnotexist']
        self.assertRaises(mod.VerbNotSupported, filtr.apply_filter,
                          self.sample_query, verb='doesnotexist')

    def test_filter_is_any(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_any(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 9)


class TestAlchemyFilter(WebTestCase):

    def setUp(self):
        self.setup_web()

        model = self.app.model
        self.sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'ggg'},
            {'name': 'foo4', 'value': 'ggg'},
            {'name': 'foo5', 'value': 'ggg'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': None},
        ]
        for setting in self.sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        self.sample_query = self.session.query(model.Setting)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop('factory', mod.AlchemyFilter)
        kwargs['model_property'] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_filter_equal(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # not filtered for null value
        filtered_query = filtr.filter_equal(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)

        # nb. by default, *is filtered* by empty string
        filtered_query = filtr.filter_equal(self.sample_query, '')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 0)

        # filtered by value
        filtered_query = filtr.filter_equal(self.sample_query, 'ggg')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 3)

    def test_filter_not_equal(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # not filtered for empty value
        filtered_query = filtr.filter_not_equal(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)

        # nb. by default, *is filtered* by empty string
        filtered_query = filtr.filter_not_equal(self.sample_query, '')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 9)

        # filtered by value
        filtered_query = filtr.filter_not_equal(self.sample_query, 'ggg')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

    def test_filter_is_null(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_null(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 1)

    def test_filter_is_not_null(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_not_null(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 8)


class TestStringAlchemyFilter(WebTestCase):

    def setUp(self):
        self.setup_web()

        model = self.app.model
        self.sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'ggg'},
            {'name': 'foo4', 'value': 'ggg'},
            {'name': 'foo5', 'value': 'ggg'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in self.sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        self.sample_query = self.session.query(model.Setting)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop('factory', mod.StringAlchemyFilter)
        kwargs['model_property'] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_filter_contains(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # not filtered for empty value
        filtered_query = filtr.filter_contains(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        filtered_query = filtr.filter_contains(self.sample_query, '')
        self.assertIs(filtered_query, self.sample_query)

        # filtered by value
        filtered_query = filtr.filter_contains(self.sample_query, 'ggg')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 3)

    def test_filter_does_not_contain(self):
        model = self.app.model
        filtr = self.make_filter(model.Setting.value)
        self.assertEqual(self.sample_query.count(), 9)

        # not filtered for empty value
        filtered_query = filtr.filter_does_not_contain(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        filtered_query = filtr.filter_does_not_contain(self.sample_query, '')
        self.assertIs(filtered_query, self.sample_query)

        # filtered by value
        filtered_query = filtr.filter_does_not_contain(self.sample_query, 'ggg')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)


class TestIntegerAlchemyFilter(WebTestCase):

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop('factory', mod.IntegerAlchemyFilter)
        kwargs['model_property'] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_coerce_value(self):
        model = self.app.model
        filtr = self.make_filter(model.Upgrade.exit_code)

        # null
        self.assertIsNone(filtr.coerce_value(None))
        self.assertIsNone(filtr.coerce_value(''))

        # typical
        self.assertEqual(filtr.coerce_value('42'), 42)
        self.assertEqual(filtr.coerce_value('-42'), -42)

        # invalid
        self.assertIsNone(filtr.coerce_value('42.12'))
        self.assertIsNone(filtr.coerce_value('bogus'))


class TestBooleanAlchemyFilter(WebTestCase):

    def setUp(self):
        self.setup_web()

        model = self.app.model
        self.sample_data = [
            {'username': 'alice',
             'prevent_edit': False,
             'active': True},
            {'username': 'bob',
             'prevent_edit': True,
             'active': True},
            {'username': 'charlie',
             'active': False,
             'prevent_edit': None},
        ]
        for user in self.sample_data:
            user = model.User(**user)
            self.session.add(user)
        self.session.commit()
        self.sample_query = self.session.query(model.User)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop('factory', mod.BooleanAlchemyFilter)
        kwargs['model_property'] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_get_verbs(self):
        model = self.app.model

        # bool field, not nullable
        filtr = self.make_filter(model.User.active,
                                 factory=mod.BooleanAlchemyFilter,
                                 nullable=False)
        self.assertFalse(hasattr(filtr, 'verbs'))
        self.assertEqual(filtr.default_verbs, ['is_true', 'is_false'])

        # by default, returns default verbs (plus 'is_any')
        self.assertEqual(filtr.get_verbs(), ['is_true', 'is_false', 'is_any'])

        # default verbs can be a callable
        filtr.default_verbs = lambda: ['foo', 'bar']
        self.assertEqual(filtr.get_verbs(), ['foo', 'bar', 'is_any'])

        # bool field, *nullable*
        filtr = self.make_filter(model.User.active,
                                 factory=mod.BooleanAlchemyFilter,
                                 nullable=True)
        self.assertFalse(hasattr(filtr, 'verbs'))
        self.assertEqual(filtr.default_verbs, ['is_true', 'is_false'])

        # effective verbs also include is_false_null
        self.assertEqual(filtr.get_verbs(), ['is_true', 'is_false', 'is_false_null',
                                             'is_null', 'is_not_null', 'is_any'])

    def test_coerce_value(self):
        model = self.app.model
        filtr = self.make_filter(model.User.active)

        self.assertIsNone(filtr.coerce_value(None))

        self.assertTrue(filtr.coerce_value(True))
        self.assertTrue(filtr.coerce_value(1))
        self.assertTrue(filtr.coerce_value('1'))

        self.assertFalse(filtr.coerce_value(False))
        self.assertFalse(filtr.coerce_value(0))
        self.assertFalse(filtr.coerce_value(''))

    def test_filter_is_true(self):
        model = self.app.model
        filtr = self.make_filter(model.User.active)
        self.assertEqual(self.sample_query.count(), 3)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_true(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

    def test_filter_is_false(self):
        model = self.app.model
        filtr = self.make_filter(model.User.active)
        self.assertEqual(self.sample_query.count(), 3)

        # nb. value None is ignored
        filtered_query = filtr.filter_is_false(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 1)

    def test_filter_is_false_null(self):
        model = self.app.model
        filtr = self.make_filter(model.User.prevent_edit)
        self.assertEqual(self.sample_query.count(), 3)

        # nb. only one account is marked with "prevent edit"
        filtered_query = filtr.filter_is_false_null(self.sample_query, None)
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)


class TheLocalThing(Base):
    __tablename__ = 'the_local_thing'
    id = sa.Column(sa.Integer(), primary_key=True, autoincrement=False)
    date = sa.Column(sa.DateTime(timezone=True), nullable=True)


class TestDateAlchemyFilter(WebTestCase):

    def setUp(self):
        self.setup_web()
        model = self.app.model

        # nb. create table for TheLocalThing
        model.Base.metadata.create_all(bind=self.session.bind)

        self.sample_data = [
            {'id': 1, 'date': datetime.date(2024, 1, 1)},
            {'id': 2, 'date': datetime.date(2024, 1, 1)},
            {'id': 3, 'date': datetime.date(2024, 3, 1)},
            {'id': 4, 'date': datetime.date(2024, 3, 1)},
            {'id': 5, 'date': None},
            {'id': 6, 'date': None},
        ]

        for thing in self.sample_data:
            thing = TheLocalThing(**thing)
            self.session.add(thing)
        self.session.commit()

        self.sample_query = self.session.query(TheLocalThing)

    def make_filter(self, model_property, **kwargs):
        factory = kwargs.pop('factory', mod.DateAlchemyFilter)
        kwargs['model_property'] = model_property
        return factory(self.request, model_property.key, **kwargs)

    def test_coerce_value(self):
        filtr = self.make_filter(TheLocalThing.date)

        # null value
        self.assertIsNone(filtr.coerce_value(None))

        # value as datetime
        value = datetime.date(2024, 1, 1)
        result = filtr.coerce_value(value)
        self.assertIs(value, result)

        # value as string
        result = filtr.coerce_value('2024-04-01')
        self.assertIsInstance(result, datetime.date)
        self.assertEqual(result, datetime.date(2024, 4, 1))

        # invalid
        result = filtr.coerce_value('thisinputisbad')
        self.assertIsNone(result)

    def test_greater_than(self):
        model = self.app.model

        filtr = self.make_filter(TheLocalThing.date)
        self.assertEqual(self.sample_query.count(), 6)

        # null value ignored
        filtered_query = filtr.filter_greater_than(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

        # value as date
        filtered_query = filtr.filter_greater_than(self.sample_query, datetime.date(2024, 2, 1))
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as string
        filtered_query = filtr.filter_greater_than(self.sample_query, '2024-02-01')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

    def test_greater_equal(self):
        model = self.app.model

        filtr = self.make_filter(TheLocalThing.date)
        self.assertEqual(self.sample_query.count(), 6)

        # null value ignored
        filtered_query = filtr.filter_greater_equal(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

        # value as date (clear of boundary)
        filtered_query = filtr.filter_greater_equal(self.sample_query, datetime.date(2024, 2, 1))
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as date (at boundary)
        filtered_query = filtr.filter_greater_equal(self.sample_query, datetime.date(2024, 3, 1))
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as string
        filtered_query = filtr.filter_greater_equal(self.sample_query, '2024-01-01')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 4)

    def test_less_than(self):
        model = self.app.model

        filtr = self.make_filter(TheLocalThing.date)
        self.assertEqual(self.sample_query.count(), 6)

        # null value ignored
        filtered_query = filtr.filter_less_than(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

        # value as date
        filtered_query = filtr.filter_less_than(self.sample_query, datetime.date(2024, 2, 1))
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as string
        filtered_query = filtr.filter_less_than(self.sample_query, '2024-04-01')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 4)

    def test_less_equal(self):
        model = self.app.model

        filtr = self.make_filter(TheLocalThing.date)
        self.assertEqual(self.sample_query.count(), 6)

        # null value ignored
        filtered_query = filtr.filter_less_equal(self.sample_query, None)
        self.assertIs(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 6)

        # value as date (clear of boundary)
        filtered_query = filtr.filter_less_equal(self.sample_query, datetime.date(2024, 2, 1))
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as date (at boundary)
        filtered_query = filtr.filter_less_equal(self.sample_query, datetime.date(2024, 3, 1))
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 2)

        # value as string
        filtered_query = filtr.filter_less_equal(self.sample_query, '2024-04-01')
        self.assertIsNot(filtered_query, self.sample_query)
        self.assertEqual(filtered_query.count(), 4)


class TestVerbNotSupported(TestCase):

    def test_basic(self):
        error = mod.VerbNotSupported('equal')
        self.assertEqual(str(error), "unknown filter verb not supported: equal")



# -*- coding: utf-8; -*-
"""
Tasks for wuttaweb
"""

import os
import shutil

from invoke import task


@task
def release(c, skip_tests=False):
    """
    Release a new version of WuttJamaican
    """
    if not skip_tests:
        c.run('pytest')

    # rebuild pkg
    if os.path.exists('dist'):
        shutil.rmtree('dist')
    if os.path.exists('WuttJamaican.egg-info'):
        shutil.rmtree('WuttJamaican.egg-info')
    c.run('python -m build --sdist')

    # upload
    c.run('twine upload dist/*')



# -*- coding: utf-8; -*-

from unittest.mock import patch

from sqlalchemy import orm

import colander

from wuttaweb.grids import Grid
from wuttaweb.views import users as mod
from wuttaweb.testing import WebTestCase


class TestUserView(WebTestCase):

    def make_view(self):
        return mod.UserView(self.request)

    def test_includeme(self):
        self.pyramid_config.include('wuttaweb.views.users')

    def test_get_query(self):
        view = self.make_view()
        query = view.get_query(session=self.session)
        self.assertIsInstance(query, orm.Query)

    def test_configure_grid(self):
        model = self.app.model
        view = self.make_view()
        grid = view.make_grid(model_class=model.User)
        self.assertFalse(grid.is_linked('person'))
        view.configure_grid(grid)
        self.assertTrue(grid.is_linked('person'))

    def test_grid_row_class(self):
        model = self.app.model
        user = model.User(username='barney', active=True)
        data = dict(user)
        view = self.make_view()

        self.assertIsNone(view.grid_row_class(user, data, 1))

        user.active = False
        self.assertEqual(view.grid_row_class(user, data, 1), 'has-background-warning')

    def test_is_editable(self):
        model = self.app.model
        view = self.make_view()

        # active user is editable
        user = model.User(username='barney', active=True)
        self.assertTrue(view.is_editable(user))

        # inactive also editable
        user = model.User(username='barney', active=False)
        self.assertTrue(view.is_editable(user))

        # but not if prevent_edit flag is set
        user = model.User(username='barney', prevent_edit=True)
        self.assertFalse(view.is_editable(user))

        # unless request user is root
        self.request.is_root = True
        self.assertTrue(view.is_editable(user))

    def test_configure_form(self):
        model = self.app.model
        person = model.Person(first_name='Barney', last_name='Rubble', full_name='Barney Rubble')
        barney = model.User(username='barney', person=person)
        self.session.add(barney)
        self.session.commit()
        view = self.make_view()

        # person replaced with first/last name when creating or editing
        with patch.object(view, 'viewing', new=True):
            form = view.make_form(model_instance=barney)
            self.assertIn('person', form)
            self.assertNotIn('first_name', form)
            self.assertNotIn('last_name', form)
            view.configure_form(form)
            self.assertIn('person', form)
            self.assertNotIn('first_name', form)
            self.assertNotIn('last_name', form)
        with patch.object(view, 'creating', new=True):
            form = view.make_form(model_instance=barney)
            self.assertIn('person', form)
            self.assertNotIn('first_name', form)
            self.assertNotIn('last_name', form)
            view.configure_form(form)
            self.assertNotIn('person', form)
            self.assertIn('first_name', form)
            self.assertIn('last_name', form)
        with patch.object(view, 'editing', new=True):
            form = view.make_form(model_instance=barney)
            self.assertIn('person', form)
            self.assertNotIn('first_name', form)
            self.assertNotIn('last_name', form)
            view.configure_form(form)
            self.assertNotIn('person', form)
            self.assertIn('first_name', form)
            self.assertIn('last_name', form)

        # first/last name have default values when editing
        with patch.object(view, 'editing', new=True):
            form = view.make_form(model_instance=barney)
            self.assertNotIn('first_name', form.defaults)
            self.assertNotIn('last_name', form.defaults)
            view.configure_form(form)
            self.assertIn('first_name', form.defaults)
            self.assertEqual(form.defaults['first_name'], 'Barney')
            self.assertIn('last_name', form.defaults)
            self.assertEqual(form.defaults['last_name'], 'Rubble')

        # password removed (always, for now)
        with patch.object(view, 'viewing', new=True):
            form = view.make_form(model_instance=barney)
            self.assertIn('password', form)
            view.configure_form(form)
            self.assertNotIn('password', form)
        with patch.object(view, 'editing', new=True):
            form = view.make_form(model_instance=barney)
            self.assertIn('password', form)
            view.configure_form(form)
            self.assertNotIn('password', form)

        # api tokens grid shown only if current user has perm
        with patch.object(view, 'viewing', new=True):
            form = view.make_form(model_instance=barney)
            self.assertIn('api_tokens', form)
            view.configure_form(form)
            self.assertNotIn('api_tokens', form)
            with patch.object(self.request, 'is_root', new=True):
                form = view.make_form(model_instance=barney)
                self.assertIn('api_tokens', form)
                view.configure_form(form)
                self.assertIn('api_tokens', form)

    def test_unique_username(self):
        model = self.app.model
        view = self.make_view()

        user = model.User(username='foo')
        self.session.add(user)
        self.session.commit()

        with patch.object(view, 'Session', return_value=self.session):

            # invalid if same username in data
            node = colander.SchemaNode(colander.String(), name='username')
            self.assertRaises(colander.Invalid, view.unique_username, node, 'foo')

            # but not if username belongs to current user
            view.editing = True
            self.request.matchdict = {'uuid': user.uuid}
            node = colander.SchemaNode(colander.String(), name='username')
            self.assertIsNone(view.unique_username(node, 'foo'))

    def test_objectify(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        view = self.make_view()

        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        others = model.Role(name="Others")
        self.session.add(others)
        barney = model.User(username='barney')
        auth.set_user_password(barney, 'testpass')
        barney.roles.append(blokes)
        self.session.add(barney)
        self.session.commit()

        with patch.object(self.request, 'matchdict', new={'uuid': barney.uuid}):
            with patch.object(view, 'editing', new=True):

                # sanity check, user is just in 'blokes' role
                self.session.refresh(barney)
                self.assertEqual(len(barney.roles), 1)
                self.assertEqual(barney.roles[0].name, "Blokes")

                # form can update user password
                self.assertTrue(auth.check_user_password(barney, 'testpass'))
                form = view.make_model_form(model_instance=barney)
                form.validated = {'username': 'barney', 'set_password': 'testpass2'}
                with patch.object(view, 'Session', return_value=self.session):
                    user = view.objectify(form)
                self.assertIs(user, barney)
                self.assertTrue(auth.check_user_password(barney, 'testpass2'))

                # form can update user roles
                form = view.make_model_form(model_instance=barney)
                form.validated = {'username': 'barney', 'roles': {others.uuid}}
                with patch.object(view, 'Session', return_value=self.session):
                    user = view.objectify(form)
                self.assertIs(user, barney)
                self.assertEqual(len(user.roles), 1)
                self.assertEqual(user.roles[0].name, "Others")

                # person is auto-created
                self.assertIsNone(barney.person)
                form = view.make_model_form(model_instance=barney)
                form.validated = {'username': 'barney', 'first_name': 'Barney', 'last_name': 'Rubble'}
                with patch.object(view, 'Session', return_value=self.session):
                    user = view.objectify(form)
                self.assertIsNotNone(barney.person)
                self.assertEqual(barney.person.first_name, 'Barney')
                self.assertEqual(barney.person.last_name, 'Rubble')
                self.assertEqual(barney.person.full_name, 'Barney Rubble')

                # person is auto-removed
                self.assertIsNotNone(barney.person)
                form = view.make_model_form(model_instance=barney)
                form.validated = {'username': 'barney', 'first_name': '', 'last_name': ''}
                with patch.object(view, 'Session', return_value=self.session):
                    user = view.objectify(form)
                self.assertIsNone(barney.person)

                # nb. re-attach the person
                barney.person = self.session.query(model.Person).one()

                # person name is updated
                self.assertEqual(barney.person.first_name, 'Barney')
                self.assertEqual(barney.person.last_name, 'Rubble')
                self.assertEqual(barney.person.full_name, 'Barney Rubble')
                form = view.make_model_form(model_instance=barney)
                form.validated = {'username': 'barney', 'first_name': 'Fred', 'last_name': 'Flintstone'}
                with patch.object(view, 'Session', return_value=self.session):
                    user = view.objectify(form)
                self.assertIsNotNone(barney.person)
                self.assertEqual(barney.person.first_name, 'Fred')
                self.assertEqual(barney.person.last_name, 'Flintstone')
                self.assertEqual(barney.person.full_name, 'Fred Flintstone')

        with patch.object(view, 'creating', new=True):

            # person is auto-created when making new user
            form = view.make_model_form()
            form.validated = {'username': 'betty', 'first_name': 'Betty', 'last_name': 'Boop'}
            with patch.object(view, 'Session', return_value=self.session):
                user = view.objectify(form)
            self.assertIsNotNone(user.person)
            self.assertEqual(user.person.first_name, 'Betty')
            self.assertEqual(user.person.last_name, 'Boop')
            self.assertEqual(user.person.full_name, 'Betty Boop')

            # nb. keep ref to last user
            last_user = user

            # person is *not* auto-created if no name provided
            form = view.make_model_form()
            form.validated = {'username': 'betty', 'first_name': '', 'last_name': ''}
            with patch.object(view, 'Session', return_value=self.session):
                user = view.objectify(form)
            self.assertIsNone(user.person)
            self.assertIsNot(user, last_user)

    def test_update_roles(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        admin = auth.get_role_administrator(self.session)
        authed = auth.get_role_authenticated(self.session)
        anon = auth.get_role_anonymous(self.session)
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        others = model.Role(name="Others")
        self.session.add(others)
        barney = model.User(username='barney')
        barney.roles.append(blokes)
        self.session.add(barney)
        self.session.commit()
        view = self.make_view()
        view.editing = True
        self.request.matchdict = {'uuid': barney.uuid}

        # no error if data is missing roles
        form = view.make_model_form(model_instance=barney)
        form.validated = {'username': 'barneyx'}
        user = view.objectify(form)
        self.assertIs(user, barney)
        self.assertEqual(barney.username, 'barneyx')

        # sanity check, user is just in 'blokes' role
        self.session.refresh(barney)
        self.assertEqual(len(barney.roles), 1)
        self.assertEqual(barney.roles[0].name, "Blokes")

        # let's test a bunch at once to ensure:
        # - user roles are updated
        # - authed / anon roles are not added
        # - admin role not added if current user is not root
        form = view.make_model_form(model_instance=barney)
        form.validated = {'username': 'barney',
                          'roles': {admin.uuid, authed.uuid, anon.uuid, others.uuid}}
        with patch.object(view, 'Session', return_value=self.session):
            user = view.objectify(form)
        self.assertIs(user, barney)
        self.assertEqual(len(user.roles), 1)
        self.assertEqual(user.roles[0].name, "Others")

        # let's test a bunch at once to ensure:
        # - user roles are updated
        # - admin role is added if current user is root
        self.request.is_root = True
        form = view.make_model_form(model_instance=barney)
        form.validated = {'username': 'barney',
                          'roles': {admin.uuid, blokes.uuid, others.uuid}}
        with patch.object(view, 'Session', return_value=self.session):
            user = view.objectify(form)
        self.assertIs(user, barney)
        self.assertEqual(len(user.roles), 3)
        role_uuids = set([role.uuid for role in user.roles])
        self.assertEqual(role_uuids, {admin.uuid, blokes.uuid, others.uuid})

        # admin role not removed if current user is not root
        self.request.is_root = False
        form = view.make_model_form(model_instance=barney)
        form.validated = {'username': 'barney',
                          'roles': {blokes.uuid, others.uuid}}
        with patch.object(view, 'Session', return_value=self.session):
            user = view.objectify(form)
        self.assertIs(user, barney)
        self.assertEqual(len(user.roles), 3)

        # admin role is removed if current user is root
        self.request.is_root = True
        form = view.make_model_form(model_instance=barney)
        form.validated = {'username': 'barney',
                          'roles': {blokes.uuid, others.uuid}}
        with patch.object(view, 'Session', return_value=self.session):
            user = view.objectify(form)
        self.assertIs(user, barney)
        self.assertEqual(len(user.roles), 2)

    def test_normalize_api_token(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        view = self.make_view()

        user = model.User(username='foo')
        self.session.add(user)
        token = auth.add_api_token(user, 'test token')
        self.session.commit()

        normal = view.normalize_api_token(token)
        self.assertIn('uuid', normal)
        self.assertEqual(normal['uuid'], token.uuid.hex)
        self.assertIn('description', normal)
        self.assertEqual(normal['description'], 'test token')
        self.assertIn('created', normal)

    def test_make_api_tokens_grid(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        view = self.make_view()

        user = model.User(username='foo')
        self.session.add(user)
        token1 = auth.add_api_token(user, 'test1')
        token2 = auth.add_api_token(user, 'test2')
        self.session.commit()

        # grid should have 2 records but no tools/actions
        grid = view.make_api_tokens_grid(user)
        self.assertIsInstance(grid, Grid)
        self.assertEqual(len(grid.data), 2)
        self.assertEqual(len(grid.tools), 0)
        self.assertEqual(len(grid.actions), 0)

        # create + delete allowed
        with patch.object(self.request, 'is_root', new=True):
            grid = view.make_api_tokens_grid(user)
            self.assertEqual(len(grid.tools), 1)
            self.assertIn('create', grid.tools)
            self.assertEqual(len(grid.actions), 1)
            self.assertEqual(grid.actions[0].key, 'delete')

    def test_add_api_token(self):
        model = self.app.model
        view = self.make_view()

        user = model.User(username='foo')
        self.session.add(user)
        self.session.commit()
        self.session.refresh(user)
        self.assertEqual(len(user.api_tokens), 0)

        with patch.object(view, 'Session', return_value=self.session):
            with patch.object(self.request, 'matchdict', new={'uuid': user.uuid}):
                with patch.object(self.request, 'json_body', create=True,
                                  new={'description': 'testing'}):
                    result = view.add_api_token()
                    self.assertEqual(len(user.api_tokens), 1)
                    token = user.api_tokens[0]
                    self.assertEqual(result['token_string'], token.token_string)
                    self.assertEqual(result['description'], 'testing')

    def test_delete_api_token(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        view = self.make_view()

        user = model.User(username='foo')
        self.session.add(user)
        token1 = auth.add_api_token(user, 'test1')
        token2 = auth.add_api_token(user, 'test2')
        self.session.commit()
        self.session.refresh(user)
        self.assertEqual(len(user.api_tokens), 2)

        with patch.object(view, 'Session', return_value=self.session):
            with patch.object(self.request, 'matchdict', new={'uuid': user.uuid}):

                # normal behavior
                with patch.object(self.request, 'json_body', create=True,
                                  new={'uuid': token1.uuid.hex}):
                    result = view.delete_api_token()
                    self.assertEqual(result, {})
                    self.session.refresh(user)
                    self.assertEqual(len(user.api_tokens), 1)
                    token = user.api_tokens[0]
                    self.assertIs(token, token2)

                # token for wrong user
                user2 = model.User(username='bar')
                self.session.add(user2)
                token3 = auth.add_api_token(user2, 'test3')
                self.session.commit()
                with patch.object(self.request, 'json_body', create=True,
                                  new={'uuid': token3.uuid.hex}):
                    result = view.delete_api_token()
                    self.assertEqual(result, {'error': "API token not found"})

                # token not found
                with patch.object(self.request, 'json_body', create=True,
                                  new={'uuid': self.app.make_true_uuid().hex}):
                    result = view.delete_api_token()
                    self.assertEqual(result, {'error': "API token not found"})



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for users
"""

import colander

from wuttjamaican.db.model import User
from wuttaweb.views import MasterView
from wuttaweb.forms import widgets
from wuttaweb.forms.schema import PersonRef, RoleRefs


class UserView(MasterView):
    """
    Master view for users.

    Default route prefix is ``users``.

    Notable URLs provided by this class:

    * ``/users/``
    * ``/users/new``
    * ``/users/XXX``
    * ``/users/XXX/edit``
    * ``/users/XXX/delete``
    """
    model_class = User

    labels = {
        'api_tokens': "API Tokens",
    }

    grid_columns = [
        'username',
        'person',
        'active',
    ]

    filter_defaults = {
        'username': {'active': True},
        'active': {'active': True, 'verb': 'is_true'},
    }
    sort_defaults = 'username'

    form_fields = [
        'username',
        'person',
        'active',
        'prevent_edit',
        'roles',
        'api_tokens',
    ]

    def get_query(self, session=None):
        """ """
        query = super().get_query(session=session)

        # nb. always join Person
        model = self.app.model
        query = query.outerjoin(model.Person)

        return query

    def configure_grid(self, g):
        """ """
        super().configure_grid(g)
        model = self.app.model

        # never show these
        g.remove('person_uuid',
                 'role_refs',
                 'password')

        # username
        g.set_link('username')

        # person
        g.set_link('person')
        g.set_sorter('person', model.Person.full_name)
        g.set_filter('person', model.Person.full_name,
                     label="Person Full Name")

    def grid_row_class(self, user, data, i):
        """ """
        if not user.active:
            return 'has-background-warning'

    def is_editable(self, user):
        """ """

        # only root can edit certain users
        if user.prevent_edit and not self.request.is_root:
            return False

        return True

    def configure_form(self, f):
        """ """
        super().configure_form(f)
        user = f.model_instance

        # username
        f.set_validator('username', self.unique_username)

        # person
        if self.creating or self.editing:
            f.fields.insert_after('person', 'first_name')
            f.set_required('first_name', False)
            f.fields.insert_after('first_name', 'last_name')
            f.set_required('last_name', False)
            f.remove('person')
            if self.editing:
                person = user.person
                if person:
                    f.set_default('first_name', person.first_name)
                    f.set_default('last_name', person.last_name)
        else:
            f.set_node('person', PersonRef(self.request))

        # password
        # nb. we must avoid 'password' as field name since
        # ColanderAlchemy wants to handle the raw/hashed value
        f.remove('password')
        # nb. no need for password field if readonly
        if self.creating or self.editing:
            # nb. use 'set_password' as field name
            f.append('set_password')
            f.set_required('set_password', False)
            f.set_widget('set_password', widgets.CheckedPasswordWidget())

        # roles
        f.append('roles')
        f.set_node('roles', RoleRefs(self.request))
        if not self.creating:
            f.set_default('roles', [role.uuid.hex for role in user.roles])

        # api_tokens
        if self.viewing and self.has_perm('manage_api_tokens'):
            f.set_grid('api_tokens', self.make_api_tokens_grid(user))
        else:
            f.remove('api_tokens')

    def unique_username(self, node, value):
        """ """
        model = self.app.model
        session = self.Session()

        query = session.query(model.User)\
                       .filter(model.User.username == value)

        if self.editing:
            uuid = self.request.matchdict['uuid']
            query = query.filter(model.User.uuid != uuid)

        if query.count():
            node.raise_invalid("Username must be unique")

    def objectify(self, form):
        """ """
        model = self.app.model
        auth = self.app.get_auth_handler()
        data = form.validated

        # normal logic first
        user = super().objectify(form)

        # maybe update person name
        if 'first_name' in form or 'last_name' in form:
            first_name = data.get('first_name')
            last_name = data.get('last_name')
            if self.creating and (first_name or last_name):
                user.person = auth.make_person(first_name=first_name, last_name=last_name)
            elif self.editing:
                if first_name or last_name:
                    if user.person:
                        person = user.person
                        if 'first_name' in form:
                            person.first_name = first_name
                        if 'last_name' in form:
                            person.last_name = last_name
                        person.full_name = self.app.make_full_name(person.first_name,
                                                                   person.last_name)
                    else:
                        user.person = auth.make_person(first_name=first_name, last_name=last_name)
                elif user.person:
                    user.person = None

        # maybe set user password
        if 'set_password' in form and data.get('set_password'):
            auth.set_user_password(user, data['set_password'])

        # update roles for user
        # TODO
        # if self.has_perm('edit_roles'):
        self.update_roles(user, form)

        return user

    def update_roles(self, user, form):
        """ """
        # TODO
        # if not self.has_perm('edit_roles'):
        #     return
        data = form.validated
        if 'roles' not in data:
            return

        model = self.app.model
        session = self.Session()
        auth = self.app.get_auth_handler()

        old_roles = set([role.uuid for role in user.roles])
        new_roles = data['roles']

        admin = auth.get_role_administrator(session)
        ignored = {
            auth.get_role_authenticated(session).uuid,
            auth.get_role_anonymous(session).uuid,
        }

        # add any new roles for the user, taking care to avoid certain
        # unwanted operations for built-in roles
        for uuid in new_roles:
            if uuid in ignored:
                continue
            if uuid in old_roles:
                continue
            if uuid == admin.uuid and not self.request.is_root:
                continue
            role = session.get(model.Role, uuid)
            user.roles.append(role)

        # remove any roles which were *not* specified, taking care to
        # avoid certain unwanted operations for built-in roles
        for uuid in old_roles:
            if uuid in new_roles:
                continue
            if uuid == admin.uuid and not self.request.is_root:
                continue
            role = session.get(model.Role, uuid)
            user.roles.remove(role)

    def make_api_tokens_grid(self, user):
        """
        Make and return the grid for the API Tokens field.

        This is only shown when current user has permission to manage
        API tokens for other users.

        :rtype: :class:`~wuttaweb.grids.base.Grid`
        """
        model = self.app.model
        route_prefix = self.get_route_prefix()

        grid = self.make_grid(key=f'{route_prefix}.view.api_tokens',
                              data=[self.normalize_api_token(t) for t in user.api_tokens],
                              columns=[
                                  'description',
                                  'created',
                              ],
                              sortable=True,
                              sort_on_backend=False,
                              sort_defaults=[('created', 'desc')])

        if self.has_perm('manage_api_tokens'):

            # create token
            button = self.make_button("New", primary=True, icon_left='plus', **{'@click': "$emit('new-token')"})
            grid.add_tool(button, key='create')

            # delete token
            grid.add_action('delete', url='#', icon='trash', link_class='has-text-danger', click_handler="$emit('delete-token', props.row)")

        return grid

    def normalize_api_token(self, token):
        """ """
        return {
            'uuid': token.uuid.hex,
            'description': token.description,
            'created': self.app.render_datetime(token.created),
        }

    def add_api_token(self):
        """
        AJAX view for adding a new user API token.

        This calls
        :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.add_api_token()`
        for the creation logic.
        """
        session = self.Session()
        auth = self.app.get_auth_handler()
        user = self.get_instance()
        data = self.request.json_body

        token = auth.add_api_token(user, data['description'])
        session.flush()
        session.refresh(token)

        result = self.normalize_api_token(token)
        result['token_string'] = token.token_string
        result['_action_url_delete'] = '#'
        return result

    def delete_api_token(self):
        """
        AJAX view for deleting a user API token.

        This calls
        :meth:`wuttjamaican:wuttjamaican.auth.AuthHandler.delete_api_token()`
        for the deletion logic.
        """
        model = self.app.model
        session = self.Session()
        auth = self.app.get_auth_handler()
        user = self.get_instance()
        data = self.request.json_body

        token = session.get(model.UserAPIToken, data['uuid'])
        if not token:
            return {'error': "API token not found"}

        if token.user is not user:
            return {'error': "API token not found"}

        auth.delete_api_token(token)
        return {}

    @classmethod
    def defaults(cls, config):
        """ """

        # nb. User may come from custom model
        wutta_config = config.registry.settings['wutta_config']
        app = wutta_config.get_app()
        cls.model_class = app.model.User

        cls._user_defaults(config)
        cls._defaults(config)

    @classmethod
    def _user_defaults(cls, config):
        """
        Provide extra default configuration for the User master view.
        """
        route_prefix = cls.get_route_prefix()
        permission_prefix = cls.get_permission_prefix()
        instance_url_prefix = cls.get_instance_url_prefix()
        model_title = cls.get_model_title()

        # manage API tokens
        config.add_wutta_permission(permission_prefix,
                                    f'{permission_prefix}.manage_api_tokens',
                                    f"Manage API tokens for any {model_title}")
        config.add_route(f'{route_prefix}.add_api_token',
                         f'{instance_url_prefix}/add-api-token',
                         request_method='POST')
        config.add_view(cls, attr='add_api_token',
                        route_name=f'{route_prefix}.add_api_token',
                        permission=f'{permission_prefix}.manage_api_tokens',
                        renderer='json')
        config.add_route(f'{route_prefix}.delete_api_token',
                         f'{instance_url_prefix}/delete-api-token',
                         request_method='POST')
        config.add_view(cls, attr='delete_api_token',
                        route_name=f'{route_prefix}.delete_api_token',
                        permission=f'{permission_prefix}.manage_api_tokens',
                        renderer='json')


def defaults(config, **kwargs):
    base = globals()

    UserView = kwargs.get('UserView', base['UserView'])
    UserView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
See also: :ref:`wutta-webapp`
"""

import os
import sys
from typing_extensions import Annotated

import typer
from pyramid.scripts import pserve

from wuttjamaican.cli import wutta_typer


@wutta_typer.command()
def webapp(
        ctx: typer.Context,
        auto_reload: Annotated[
            bool,
            typer.Option('--reload', '-r',
                         help="Auto-reload web app when files change.")] = False,
):
    """
    Run the configured web app
    """
    config = ctx.parent.wutta_config

    # we'll need config file(s) to specify for web app
    if not config.files_read:
        sys.stderr.write("no config files found!\n")
        sys.exit(1)

    runner = config.get(f'{config.appname}.web.app.runner', default='pserve')
    if runner == 'pserve':

        # run pserve
        argv = ['pserve', f'file+ini:{config.files_read[0]}']
        if ctx.params['auto_reload']:
            argv.append('--reload')
        pserve.main(argv=argv)

    elif runner == 'uvicorn':

        import uvicorn

        # need service details from config
        spec = config.require(f'{config.appname}.web.app.spec')
        kw = {
            'host': config.get(f'{config.appname}.web.app.host', default='127.0.0.1'),
            'port': config.get_int(f'{config.appname}.web.app.port', default=8000),
            'reload': ctx.params['auto_reload'],
            'reload_dirs': config.get_list(f'{config.appname}.web.app.reload_dirs'),
            'factory': config.get_bool(f'{config.appname}.web.app.factory', default=False),
            'interface': config.get(f'{config.appname}.web.app.interface', default='auto'),
            'root_path': config.get(f'{config.appname}.web.app.root_path', default=''),
        }

        # also must inject our config files to env, since there is no
        # other way to specify when running via uvicorn
        os.environ['WUTTA_CONFIG_FILES'] = os.pathsep.join(config.files_read)

        # run uvicorn
        uvicorn.run(spec, **kw)

    else:
        sys.stderr.write(f"unknown web app runner: {runner}\n")
        sys.exit(2)



# -*- coding: utf-8; -*-

import datetime
import decimal
from unittest.mock import patch

import colander
import deform
from pyramid import testing

from wuttaweb import grids
from wuttaweb.forms import widgets as mod
from wuttaweb.forms import schema
from wuttaweb.forms.schema import (FileDownload, PersonRef, RoleRefs, Permissions,
                                   WuttaDateTime, EmailRecipients)
from wuttaweb.testing import WebTestCase


class TestObjectRefWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault('renderer', deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def make_widget(self, **kwargs):
        return mod.ObjectRefWidget(self.request, **kwargs)

    def test_serialize(self):
        model = self.app.model
        person = model.Person(full_name="Betty Boop")
        self.session.add(person)
        self.session.commit()

        with patch.object(schema, 'Session', return_value=self.session):

            # standard (editable)
            node = colander.SchemaNode(PersonRef(self.request))
            widget = self.make_widget()
            field = self.make_field(node)
            html = widget.serialize(field, person.uuid)
            self.assertIn('<b-select ', html)

            # readonly
            node = colander.SchemaNode(PersonRef(self.request))
            node.model_instance = person
            widget = self.make_widget()
            field = self.make_field(node)
            html = widget.serialize(field, person.uuid, readonly=True)
            self.assertIn('Betty Boop', html)
            self.assertNotIn('<a', html)

            # with hyperlink
            node = colander.SchemaNode(PersonRef(self.request))
            node.model_instance = person
            widget = self.make_widget(url=lambda p: '/foo')
            field = self.make_field(node)
            html = widget.serialize(field, person.uuid, readonly=True)
            self.assertIn('Betty Boop', html)
            self.assertIn('<a', html)
            self.assertIn('href="/foo"', html)

    def test_get_template_values(self):
        model = self.app.model
        person = model.Person(full_name="Betty Boop")
        self.session.add(person)
        self.session.commit()

        with patch.object(schema, 'Session', return_value=self.session):

            # standard
            node = colander.SchemaNode(PersonRef(self.request))
            widget = self.make_widget()
            field = self.make_field(node)
            values = widget.get_template_values(field, person.uuid, {})
            self.assertIn('cstruct', values)
            self.assertNotIn('url', values)

            # readonly w/ empty option
            node = colander.SchemaNode(PersonRef(self.request,
                                                 empty_option=('_empty_', '(empty)')))
            widget = self.make_widget(readonly=True, url=lambda obj: '/foo')
            field = self.make_field(node)
            values = widget.get_template_values(field, '_empty_', {})
            self.assertIn('cstruct', values)
            self.assertNotIn('url', values)


class TestWuttaDateWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault('renderer', deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def make_widget(self, **kwargs):
        return mod.WuttaDateWidget(self.request, **kwargs)

    def test_serialize(self):
        node = colander.SchemaNode(colander.Date())
        field = self.make_field(node)

        # first try normal date
        widget = self.make_widget()
        dt = datetime.date(2025, 1, 15)

        # editable widget has normal picker html
        result = widget.serialize(field, str(dt))
        self.assertIn('<wutta-datepicker', result)

        # readonly is rendered per app convention
        result = widget.serialize(field, str(dt), readonly=True)
        self.assertEqual(result, '2025-01-15')

        # now try again with datetime
        widget = self.make_widget()
        dt = datetime.datetime(2025, 1, 15, 8, 35)

        # editable widget has normal picker html
        result = widget.serialize(field, str(dt))
        self.assertIn('<wutta-datepicker', result)

        # readonly is rendered per app convention
        result = widget.serialize(field, str(dt), readonly=True)
        self.assertEqual(result, '2025-01-15')


class TestWuttaDateTimeWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault('renderer', deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def make_widget(self, **kwargs):
        return mod.WuttaDateTimeWidget(self.request, **kwargs)

    def test_serialize(self):
        node = colander.SchemaNode(WuttaDateTime())
        field = self.make_field(node)
        widget = self.make_widget()
        dt = datetime.datetime(2024, 12, 12, 13, 49, tzinfo=datetime.timezone.utc)

        # editable widget has normal picker html
        result = widget.serialize(field, str(dt))
        self.assertIn('<wutta-datepicker', result)

        # readonly is rendered per app convention
        result = widget.serialize(field, str(dt), readonly=True)
        self.assertEqual(result, '2024-12-12 13:49+0000')


class TestWuttaMoneyInputWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault('renderer', deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def make_widget(self, **kwargs):
        return mod.WuttaMoneyInputWidget(self.request, **kwargs)

    def test_serialize(self):
        node = colander.SchemaNode(schema.WuttaMoney(self.request))
        field = self.make_field(node)
        widget = self.make_widget()
        amount = decimal.Decimal('12.34')

        # editable widget has normal text input
        result = widget.serialize(field, str(amount))
        self.assertIn('<b-input', result)

        # readonly is rendered per app convention
        result = widget.serialize(field, str(amount), readonly=True)
        self.assertEqual(result, '<span>$12.34</span>')

        # readonly w/ null value
        result = widget.serialize(field, None, readonly=True)
        self.assertEqual(result, '<span></span>')


class TestFileDownloadWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault('renderer', deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def test_serialize(self):

        # nb. we let the field construct the widget via our type
        # (nb. at first we do not provide a url)
        node = colander.SchemaNode(FileDownload(self.request))
        field = self.make_field(node)
        widget = field.widget

        # null value
        html = widget.serialize(field, None, readonly=True)
        self.assertNotIn('<a ', html)
        self.assertIn('<span>', html)

        # path to nonexistent file
        html = widget.serialize(field, '/this/path/does/not/exist', readonly=True)
        self.assertNotIn('<a ', html)
        self.assertIn('<span>', html)

        # path to actual file
        datfile = self.write_file('data.txt', "hello\n" * 1000)
        html = widget.serialize(field, datfile, readonly=True)
        self.assertNotIn('<a ', html)
        self.assertIn('<span>', html)
        self.assertIn('data.txt', html)
        self.assertIn('kB)', html)

        # path to file, w/ url
        node = colander.SchemaNode(FileDownload(self.request, url='/download/blarg'))
        field = self.make_field(node)
        widget = field.widget
        html = widget.serialize(field, datfile, readonly=True)
        self.assertNotIn('<span>', html)
        self.assertIn('<a href="/download/blarg">', html)
        self.assertIn('data.txt', html)
        self.assertIn('kB)', html)

        # nb. same readonly output even if we ask for editable
        html2 = widget.serialize(field, datfile, readonly=False)
        self.assertEqual(html2, html)


class TestGridWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault('renderer', deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def test_serialize(self):
        grid = grids.Grid(self.request,
                          columns=['foo', 'bar'],
                          data=[{'foo': 1, 'bar': 2}, {'foo': 3, 'bar': 4}])

        node = colander.SchemaNode(colander.String())
        widget = mod.GridWidget(self.request, grid)
        field = self.make_field(node)

        # readonly works okay
        html = widget.serialize(field, None, readonly=True)
        self.assertIn('<b-table ', html)

        # but otherwise, error
        self.assertRaises(NotImplementedError, widget.serialize, field, None)


class TestRoleRefsWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault('renderer', deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def test_serialize(self):
        self.pyramid_config.add_route('roles.view', '/roles/{uuid}')
        model = self.app.model
        auth = self.app.get_auth_handler()
        admin = auth.get_role_administrator(self.session)
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        self.session.commit()

        # nb. we let the field construct the widget via our type
        with patch.object(schema, 'Session', return_value=self.session):
            node = colander.SchemaNode(RoleRefs(self.request))
            field = self.make_field(node)
            widget = field.widget

            # readonly values list includes admin
            html = widget.serialize(field, {admin.uuid, blokes.uuid}, readonly=True)
            self.assertIn(admin.name, html)
            self.assertIn(blokes.name, html)

            # editable values list *excludes* admin (by default)
            html = widget.serialize(field, {admin.uuid, blokes.uuid})
            self.assertNotIn(str(admin.uuid.hex), html)
            self.assertIn(str(blokes.uuid.hex), html)

            # but admin is included for root user
            self.request.is_root = True
            node = colander.SchemaNode(RoleRefs(self.request))
            field = self.make_field(node)
            widget = field.widget
            html = widget.serialize(field, {admin.uuid, blokes.uuid})
            self.assertIn(str(admin.uuid.hex), html)
            self.assertIn(str(blokes.uuid.hex), html)


class TestPermissionsWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault('renderer', deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def test_serialize(self):
        permissions = {
            'widgets': {
                'label': "Widgets",
                'perms': {
                    'widgets.polish': {
                        'label': "Polish the widgets",
                    },
                },
            },
        }

        # nb. we let the field construct the widget via our type
        node = colander.SchemaNode(Permissions(self.request, permissions))
        field = self.make_field(node)
        widget = field.widget

        # readonly output does *not* include the perm by default
        html = widget.serialize(field, set(), readonly=True)
        self.assertNotIn("Polish the widgets", html)

        # readonly output includes the perm if set
        html = widget.serialize(field, {'widgets.polish'}, readonly=True)
        self.assertIn("Polish the widgets", html)

        # editable output always includes the perm
        html = widget.serialize(field, set())
        self.assertIn("Polish the widgets", html)


class TestEmailRecipientsWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault('renderer', deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def test_serialize(self):
        node = colander.SchemaNode(EmailRecipients())
        field = self.make_field(node)
        widget = mod.EmailRecipientsWidget()

        recips = [
            'alice@example.com',
            'bob@example.com',
        ]
        recips_str = ', '.join(recips)

        # readonly
        result = widget.serialize(field, recips_str, readonly=True)
        self.assertIn('<ul>', result)
        self.assertIn('<li>alice@example.com</li>', result)

        # editable
        result = widget.serialize(field, recips_str)
        self.assertIn('<b-input', result)
        self.assertIn('type="textarea"', result)

    def test_deserialize(self):
        node = colander.SchemaNode(EmailRecipients())
        field = self.make_field(node)
        widget = mod.EmailRecipientsWidget()

        recips = [
            'alice@example.com',
            'bob@example.com',
        ]
        recips_str = ', '.join(recips)

        # values
        result = widget.deserialize(field, recips_str)
        self.assertEqual(result, recips_str)

        # null
        result = widget.deserialize(field, colander.null)
        self.assertIs(result, colander.null)


class TestBatchIdWidget(WebTestCase):

    def make_field(self, node, **kwargs):
        # TODO: not sure why default renderer is in use even though
        # pyramid_deform was included in setup?  but this works..
        kwargs.setdefault('renderer', deform.Form.default_renderer)
        return deform.Field(node, **kwargs)

    def test_serialize(self):
        node = colander.SchemaNode(colander.Integer())
        field = self.make_field(node)
        widget = mod.BatchIdWidget()

        result = widget.serialize(field, colander.null)
        self.assertIs(result, colander.null)

        result = widget.serialize(field, 42)
        self.assertEqual(result, '00000042')



# -*- coding: utf-8; -*-

from wuttjamaican.conf import WuttaConfig
from wuttjamaican.testing import FileConfigTestCase
from wuttaweb.menus import MenuHandler


class DataTestCase(FileConfigTestCase):
    """
    Base class for test suites requiring a full (typical) database.
    """

    def setUp(self):
        self.setup_db()

    def setup_db(self):
        self.setup_files()
        self.config = WuttaConfig(defaults={
            'wutta.db.default.url': 'sqlite://',
        })
        self.app = self.config.get_app()

        # init db
        model = self.app.model
        model.Base.metadata.create_all(bind=self.config.appdb_engine)
        self.session = self.app.make_session()

    def tearDown(self):
        self.teardown_db()

    def teardown_db(self):
        self.teardown_files()


class NullMenuHandler(MenuHandler):
    """
    Dummy menu handler for testing.
    """
    def make_menus(self, request, **kwargs):
        return []



# -*- coding: utf-8; -*-

from unittest import TestCase
from unittest.mock import patch

from wuttjamaican.testing import FileTestCase, ConfigTestCase, DataTestCase

from asgiref.wsgi import WsgiToAsgi
from pyramid.config import Configurator
from pyramid.router import Router

from wuttaweb import app as mod
from wuttjamaican.conf import WuttaConfig
from wuttjamaican.app import AppHandler
from wuttjamaican.util import resource_path


class TestWebAppProvider(TestCase):

    def test_basic(self):
        # nb. just normal usage here, confirm it does the one thing we
        # need it to..
        config = WuttaConfig()
        app = config.get_app()
        handler = app.get_web_handler()


class TestMakeWuttaConfig(FileTestCase):

    def test_config_path_required(self):

        # settings must define config path, else error
        settings = {}
        self.assertRaises(ValueError, mod.make_wutta_config, settings)

    def test_basic(self):

        # mock path to config file
        myconf = self.write_file('my.conf', '')
        settings = {'wutta.config': myconf}

        # can make a config okay
        config = mod.make_wutta_config(settings)

        # and that config is also stored in settings
        self.assertIn('wutta_config', settings)
        self.assertIs(settings['wutta_config'], config)


class TestMakePyramidConfig(DataTestCase):

    def test_basic(self):
        with patch.object(AppHandler, 'make_session', return_value=self.session):
            settings = {'wutta_config': self.config}
            config = mod.make_pyramid_config(settings)
            self.assertIsInstance(config, Configurator)
            self.assertEqual(settings['wuttaweb.theme'], 'default')


class TestMain(DataTestCase):

    def test_basic(self):
        with patch.object(AppHandler, 'make_session', return_value=self.session):
            global_config = None
            myconf = self.write_file('my.conf', '')
            settings = {'wutta.config': myconf}
            app = mod.main(global_config, **settings)
            self.assertIsInstance(app, Router)


def mock_main(global_config, **settings):

    wutta_config = mod.make_wutta_config(settings)
    pyramid_config = mod.make_pyramid_config(settings)

    pyramid_config.include('wuttaweb.static')
    pyramid_config.include('wuttaweb.subscribers')
    pyramid_config.include('wuttaweb.views')

    return pyramid_config.make_wsgi_app()


class TestMakeWsgiApp(DataTestCase):

    def test_with_callable(self):
        with patch.object(self.app, 'make_session', return_value=self.session):

            # specify config
            wsgi = mod.make_wsgi_app(mock_main, config=self.config)
            self.assertIsInstance(wsgi, Router)

            # auto config
            with patch.object(mod, 'make_config', return_value=self.config):
                wsgi = mod.make_wsgi_app(mock_main)
                self.assertIsInstance(wsgi, Router)

    def test_with_spec(self):

        # specify config
        wsgi = mod.make_wsgi_app('tests.test_app:mock_main', config=self.config)
        self.assertIsInstance(wsgi, Router)

        # auto config
        with patch.object(mod, 'make_config', return_value=self.config):
            wsgi = mod.make_wsgi_app('tests.test_app:mock_main')
            self.assertIsInstance(wsgi, Router)

    def test_invalid(self):
        self.assertRaises(ValueError, mod.make_wsgi_app, 42, config=self.config)


class TestMakeAsgiApp(DataTestCase):

    def test_with_callable(self):
        with patch.object(self.app, 'make_session', return_value=self.session):

            # specify config
            asgi = mod.make_asgi_app(mock_main, config=self.config)
            self.assertIsInstance(asgi, WsgiToAsgi)

            # auto config
            with patch.object(mod, 'make_config', return_value=self.config):
                asgi = mod.make_asgi_app(mock_main)
                self.assertIsInstance(asgi, WsgiToAsgi)

    def test_with_spec(self):
        with patch.object(self.app, 'make_session', return_value=self.session):

            # specify config
            asgi = mod.make_asgi_app('tests.test_app:mock_main', config=self.config)
            self.assertIsInstance(asgi, WsgiToAsgi)

            # auto config
            with patch.object(mod, 'make_config', return_value=self.config):
                asgi = mod.make_asgi_app('tests.test_app:mock_main')
                self.assertIsInstance(asgi, WsgiToAsgi)

    def test_invalid(self):
        self.assertRaises(ValueError, mod.make_asgi_app, 42, config=self.config)


class TestEstablishTheme(DataTestCase):

    def test_default(self):
        settings = {
            'wutta_config': self.config,
            'mako.directories': ['wuttaweb:templates'],
        }
        mod.establish_theme(settings)
        self.assertEqual(settings['wuttaweb.theme'], 'default')
        self.assertEqual(settings['mako.directories'], [
            resource_path('wuttaweb:templates/themes/default'),
            'wuttaweb:templates',
        ])

    def test_mako_dirs_as_string(self):
        settings = {
            'wutta_config': self.config,
            'mako.directories': 'wuttaweb:templates',
        }
        mod.establish_theme(settings)
        self.assertEqual(settings['wuttaweb.theme'], 'default')
        self.assertEqual(settings['mako.directories'], [
            resource_path('wuttaweb:templates/themes/default'),
            'wuttaweb:templates',
        ])

    def test_butterfly(self):
        settings = {
            'wutta_config': self.config,
            'mako.directories': 'wuttaweb:templates',
        }
        self.app.save_setting(self.session, 'wuttaweb.theme', 'butterfly')
        self.session.commit()
        mod.establish_theme(settings)
        self.assertEqual(settings['wuttaweb.theme'], 'butterfly')
        self.assertEqual(settings['mako.directories'], [
            resource_path('wuttaweb:templates/themes/butterfly'),
            'wuttaweb:templates',
        ])

    def test_custom(self):
        settings = {
            'wutta_config': self.config,
            'mako.directories': 'wuttaweb:templates',
        }
        self.config.setdefault('wuttaweb.themes.keys', 'anotherone')
        self.app.save_setting(self.session, 'wuttaweb.theme', 'anotherone')
        self.session.commit()
        mod.establish_theme(settings)
        self.assertEqual(settings['wuttaweb.theme'], 'anotherone')
        self.assertEqual(settings['mako.directories'], [
            resource_path('wuttaweb:templates/themes/anotherone'),
            'wuttaweb:templates',
        ])



# -*- coding: utf-8; -*-

from unittest import TestCase
from unittest.mock import patch, MagicMock

from wuttaweb import menus as mod
from wuttaweb.testing import WebTestCase


class TestMenuHandler(WebTestCase):

    def setUp(self):
        self.setup_web()
        self.handler = mod.MenuHandler(self.config)

    def test_make_admin_menu(self):

        # no people entry by default
        menu = self.handler.make_admin_menu(self.request)
        self.assertIsInstance(menu, dict)
        routes = [item.get('route') for item in menu['items']]
        self.assertNotIn('people', routes)

        # but we can request it
        menu = self.handler.make_admin_menu(self.request, include_people=True)
        routes = [item.get('route') for item in menu['items']]
        self.assertIn('people', routes)

    def test_make_menus(self):
        menus = self.handler.make_menus(self.request)
        self.assertIsInstance(menus, list)

    def test_is_allowed(self):
        model = self.app.model
        auth = self.app.get_auth_handler()

        # user with perms
        barney = model.User(username='barney')
        self.session.add(barney)
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        barney.roles.append(blokes)
        auth.grant_permission(blokes, 'appinfo.list')
        self.request.user = barney

        # perm not granted to user
        item = {'perm': 'appinfo.configure'}
        self.assertFalse(self.handler._is_allowed(self.request, item))

        # perm *is* granted to user
        item = {'perm': 'appinfo.list'}
        self.assertTrue(self.handler._is_allowed(self.request, item))

        # perm not required
        item = {}
        self.assertTrue(self.handler._is_allowed(self.request, item))

    def test_mark_allowed(self):

        def make_menus():
            return [
                {
                    'type': 'menu',
                    'items': [
                        {'title': "Foo", 'url': '#'},
                        {'title': "Bar", 'url': '#'},
                    ],
                },
            ]

        mock_is_allowed = MagicMock()
        with patch.object(self.handler, '_is_allowed', new=mock_is_allowed):

            # all should be allowed
            mock_is_allowed.return_value = True
            menus = make_menus()
            self.handler._mark_allowed(self.request, menus)
            menu = menus[0]
            self.assertTrue(menu['allowed'])
            foo, bar = menu['items']
            self.assertTrue(foo['allowed'])
            self.assertTrue(bar['allowed'])

            # none should be allowed
            mock_is_allowed.return_value = False
            menus = make_menus()
            self.handler._mark_allowed(self.request, menus)
            menu = menus[0]
            self.assertFalse(menu['allowed'])
            foo, bar = menu['items']
            self.assertFalse(foo['allowed'])
            self.assertFalse(bar['allowed'])

    def test_mark_allowed_submenu(self):

        def make_menus():
            return [
                {
                    'type': 'menu',
                    'items': [
                        {'title': "Foo", 'url': '#'},
                        {
                            'type': 'menu',
                            'items': [
                                {'title': "Bar", 'url': '#'},
                            ],
                        },
                    ],
                },
            ]

        mock_is_allowed = MagicMock()
        with patch.object(self.handler, '_is_allowed', new=mock_is_allowed):

            # all should be allowed
            mock_is_allowed.return_value = True
            menus = make_menus()
            self.handler._mark_allowed(self.request, menus)
            menu = menus[0]
            self.assertTrue(menu['allowed'])
            foo, submenu = menu['items']
            self.assertTrue(foo['allowed'])
            self.assertTrue(submenu['allowed'])
            subitem = submenu['items'][0]
            self.assertTrue(subitem['allowed'])

            # none should be allowed
            mock_is_allowed.return_value = False
            menus = make_menus()
            self.handler._mark_allowed(self.request, menus)
            menu = menus[0]
            self.assertFalse(menu['allowed'])
            foo, submenu = menu['items']
            self.assertFalse(foo['allowed'])
            self.assertFalse(submenu['allowed'])
            subitem = submenu['items'][0]
            self.assertFalse(subitem['allowed'])

    def test_make_menu_key(self):
        self.assertEqual(self.handler._make_menu_key('foo'), 'foo')
        self.assertEqual(self.handler._make_menu_key('FooBar'), 'foobar')
        self.assertEqual(self.handler._make_menu_key('Foo - $#Bar'), 'foobar')
        self.assertEqual(self.handler._make_menu_key('Foo__Bar'), 'foo__bar')

    def test_make_menu_entry_item(self):
        item = {'title': "Foo", 'url': '#'}
        entry = self.handler._make_menu_entry(self.request, item)
        self.assertEqual(entry['type'], 'item')
        self.assertEqual(entry['title'], "Foo")
        self.assertEqual(entry['url'], '#')
        self.assertTrue(entry['is_link'])

    def test_make_menu_entry_item_with_no_url(self):
        item = {'title': "Foo"}
        entry = self.handler._make_menu_entry(self.request, item)
        self.assertEqual(entry['type'], 'item')
        self.assertEqual(entry['title'], "Foo")
        self.assertNotIn('url', entry)
        # nb. still sets is_link = True; basically it's <a> with no href
        self.assertTrue(entry['is_link'])

    def test_make_menu_entry_item_with_known_route(self):
        item = {'title': "Foo", 'route': 'home'}
        with patch.object(self.request, 'route_url', return_value='/something'):
            entry = self.handler._make_menu_entry(self.request, item)
            self.assertEqual(entry['type'], 'item')
            self.assertEqual(entry['url'], '/something')
            self.assertTrue(entry['is_link'])

    def test_make_menu_entry_item_with_unknown_route(self):
        item = {'title': "Foo", 'route': 'home'}
        with patch.object(self.request, 'route_url', side_effect=KeyError):
            entry = self.handler._make_menu_entry(self.request, item)
            self.assertEqual(entry['type'], 'item')
            # nb. fake url is used, based on (bad) route name
            self.assertEqual(entry['url'], 'home')
            self.assertTrue(entry['is_link'])

    def test_make_menu_entry_sep(self):
        item = {'type': 'sep'}
        entry = self.handler._make_menu_entry(self.request, item)
        self.assertEqual(entry['type'], 'sep')
        self.assertTrue(entry['is_sep'])
        self.assertFalse(entry['is_menu'])

    def test_make_raw_menus(self):
        # minimal test to ensure it calls the other method
        with patch.object(self.handler, 'make_menus') as make_menus:
            self.handler._make_raw_menus(self.request, foo='bar')
            make_menus.assert_called_once_with(self.request, foo='bar')

    def test_do_make_menus_prune_unallowed_item(self):
        test_menus = [
            {
                'title': "First Menu",
                'type': 'menu',
                'items': [
                    {'title': "Foo", 'url': '#'},
                    {'title': "Bar", 'url': '#'},
                ],
            },
        ]

        def is_allowed(request, item):
            if item.get('title') == 'Bar':
                return False
            return True

        with patch.object(self.handler, 'make_menus', return_value=test_menus):
            with patch.object(self.handler, '_is_allowed', side_effect=is_allowed):
                menus = self.handler.do_make_menus(self.request)

                # Foo remains but Bar is pruned
                menu = menus[0]
                self.assertEqual(len(menu['items']), 1)
                item = menu['items'][0]
                self.assertEqual(item['title'], 'Foo')

    def test_do_make_menus_prune_unallowed_menu(self):
        test_menus = [
            {
                'title': "First Menu",
                'type': 'menu',
                'items': [
                    {'title': "Foo", 'url': '#'},
                    {'title': "Bar", 'url': '#'},
                ],
            },
            {
                'title': "Second Menu",
                'type': 'menu',
                'items': [
                    {'title': "Baz", 'url': '#'},
                ],
            },
        ]

        def is_allowed(request, item):
            if item.get('title') == 'Baz':
                return True
            return False

        with patch.object(self.handler, 'make_menus', return_value=test_menus):
            with patch.object(self.handler, '_is_allowed', side_effect=is_allowed):
                menus = self.handler.do_make_menus(self.request)

                # Second/Baz remains but First/Foo/Bar are pruned
                self.assertEqual(len(menus), 1)
                menu = menus[0]
                self.assertEqual(menu['title'], 'Second Menu')
                self.assertEqual(len(menu['items']), 1)
                item = menu['items'][0]
                self.assertEqual(item['title'], 'Baz')

    def test_do_make_menus_with_top_link(self):
        test_menus = [
            {
                'title': "First Menu",
                'type': 'menu',
                'items': [
                    {'title': "Foo", 'url': '#'},
                    {'title': "Bar", 'url': '#'},
                ],
            },
            {
                'title': "Second Link",
                'type': 'link',
            },
        ]

        with patch.object(self.handler, 'make_menus', return_value=test_menus):
            with patch.object(self.handler, '_is_allowed', return_value=True):
                menus = self.handler.do_make_menus(self.request)

                # ensure top link remains
                self.assertEqual(len(menus), 2)
                menu = menus[1]
                self.assertEqual(menu['title'], "Second Link")

    def test_do_make_menus_with_trailing_sep(self):
        test_menus = [
            {
                'title': "First Menu",
                'type': 'menu',
                'items': [
                    {'title': "Foo", 'url': '#'},
                    {'title': "Bar", 'url': '#'},
                    {'type': 'sep'},
                ],
            },
        ]

        with patch.object(self.handler, 'make_menus', return_value=test_menus):
            with patch.object(self.handler, '_is_allowed', return_value=True):
                menus = self.handler.do_make_menus(self.request)

                # ensure trailing sep was pruned
                menu = menus[0]
                self.assertEqual(len(menu['items']), 2)
                foo, bar = menu['items']
                self.assertEqual(foo['title'], 'Foo')
                self.assertEqual(bar['title'], 'Bar')

    def test_do_make_menus_with_submenu(self):
        test_menus = [
            {
                'title': "First Menu",
                'type': 'menu',
                'items': [
                    {
                        'title': "First Submenu",
                        'type': 'menu',
                        'items': [
                            {'title': "Foo", 'url': '#'},
                        ],
                    },
                    {
                        'title': "Second Submenu",
                        'type': 'menu',
                        'items': [
                            {'title': "Bar", 'url': '#'},
                        ],
                    },
                ],
            },
        ]

        def is_allowed(request, item):
            if item.get('title') == 'Bar':
                return False
            return True

        with patch.object(self.handler, 'make_menus', return_value=test_menus):
            with patch.object(self.handler, '_is_allowed', side_effect=is_allowed):
                menus = self.handler.do_make_menus(self.request)

                # first submenu remains, second is pruned
                menu = menus[0]
                self.assertEqual(len(menu['items']), 1)
                submenu = menu['items'][0]
                self.assertEqual(submenu['type'], 'submenu')
                self.assertEqual(submenu['title'], 'First Submenu')
                self.assertEqual(len(submenu['items']), 1)
                item = submenu['items'][0]
                self.assertEqual(item['title'], 'Foo')



# -*- coding: utf-8; -*-

from unittest.mock import patch

from sqlalchemy import orm

from pyramid.httpexceptions import HTTPNotFound

from wuttaweb.views import people
from wuttaweb.testing import WebTestCase
from wuttaweb.forms.widgets import GridWidget
from wuttaweb.grids import Grid


class TestPersonView(WebTestCase):

    def make_view(self):
        return people.PersonView(self.request)

    def test_includeme(self):
        self.pyramid_config.include('wuttaweb.views.people')

    def test_get_query(self):
        view = self.make_view()
        query = view.get_query(session=self.session)
        self.assertIsInstance(query, orm.Query)

    def test_configure_grid(self):
        model = self.app.model
        view = self.make_view()
        grid = view.make_grid(model_class=model.Setting)
        self.assertEqual(grid.linked_columns, [])
        view.configure_grid(grid)
        self.assertIn('full_name', grid.linked_columns)

    def test_configure_form(self):
        model = self.app.model
        view = self.make_view()

        # full_name
        form = view.make_form(model_class=model.Person)
        self.assertIn('full_name', form)
        with patch.object(view, 'creating', new=True):
            view.configure_form(form)
            self.assertNotIn('full_name', form)

        # users
        person = model.Person()
        form = view.make_form(model_instance=person)
        self.assertNotIn('users', form.widgets)
        with patch.object(view, 'viewing', new=True):
            view.configure_form(form)
            self.assertIn('users', form.widgets)
            self.assertIsInstance(form.widgets['users'], GridWidget)

    def test_make_users_grid(self):
        model = self.app.model
        view = self.make_view()
        person = model.Person(full_name="John Doe")

        # basic
        grid = view.make_users_grid(person)
        self.assertIsInstance(grid, Grid)
        self.assertFalse(grid.linked_columns)
        self.assertFalse(grid.actions)

        # view + edit actions
        with patch.object(self.request, 'is_root', new=True):
            grid = view.make_users_grid(person)
            self.assertIsInstance(grid, Grid)
            self.assertIn('username', grid.linked_columns)
            self.assertEqual(len(grid.actions), 2)
            self.assertEqual(grid.actions[0].key, 'view')
            self.assertEqual(grid.actions[1].key, 'edit')

    def test_objectify(self):
        model = self.app.model
        view = self.make_view()

        # creating
        form = view.make_model_form()
        form.validated = {'first_name': 'Barney', 'last_name': 'Rubble'}
        person = view.objectify(form)
        self.assertEqual(person.full_name, 'Barney Rubble')

        # editing
        form = view.make_model_form(model_instance=person)
        form.validated = {'first_name': 'Betty', 'last_name': 'Rubble'}
        person2 = view.objectify(form)
        self.assertEqual(person2.full_name, 'Betty Rubble')
        self.assertIs(person2, person)

    def test_autocomplete_query(self):
        model = self.app.model

        person1 = model.Person(full_name="George Jones")
        self.session.add(person1)
        person2 = model.Person(full_name="George Strait")
        self.session.add(person2)
        self.session.commit()

        view = self.make_view()
        with patch.object(view, 'Session', return_value=self.session):

            # both people match
            query = view.autocomplete_query('george')
            self.assertEqual(query.count(), 2)

            # just 1 match
            query = view.autocomplete_query('jones')
            self.assertEqual(query.count(), 1)

            # no matches
            query = view.autocomplete_query('sally')
            self.assertEqual(query.count(), 0)

    def test_view_profile(self):
        self.pyramid_config.include('wuttaweb.views.common')
        self.pyramid_config.include('wuttaweb.views.auth')
        self.pyramid_config.add_route('people', '/people/')

        model = self.app.model
        person = model.Person(full_name="Barney Rubble")
        self.session.add(person)
        self.session.commit()

        # sanity check
        view = self.make_view()
        self.request.matchdict = {'uuid': person.uuid}
        response = view.view_profile(session=self.session)
        self.assertEqual(response.status_code, 200)

    def test_make_user(self):
        self.pyramid_config.include('wuttaweb.views.common')

        model = self.app.model
        person = model.Person(full_name="Barney Rubble")
        self.session.add(person)
        self.session.commit()

        # sanity check
        view = self.make_view()
        self.request.matchdict = {'uuid': person.uuid}
        response = view.make_user()
        # nb. this always redirects for now
        self.assertEqual(response.status_code, 302)



# -*- coding: utf-8; -*-

from unittest.mock import patch

from wuttjamaican.email import EmailSetting

import colander
from pyramid.httpexceptions import HTTPNotFound
from pyramid.response import Response

from wuttaweb.views import email as mod
from wuttaweb.testing import WebTestCase


class TestEmailSettingViews(WebTestCase):

    def make_view(self):
        return mod.EmailSettingView(self.request)

    def test_includeme(self):
        self.pyramid_config.include('wuttaweb.views.email')

    def test_get_grid_data(self):
        self.config.setdefault('wutta.email.default.sender', 'test@example.com')
        view = self.make_view()
        data = view.get_grid_data()
        self.assertIsInstance(data, list)
        self.assertTrue(data)   # 1+ items
        setting = data[0]
        self.assertIn('key', setting)
        self.assertIn('subject', setting)
        self.assertIn('sender', setting)
        self.assertIn('to', setting)
        self.assertIn('cc', setting)
        self.assertIn('notes', setting)

    def test_configure_grid(self):
        self.config.setdefault('wutta.email.default.sender', 'test@example.com')
        view = self.make_view()
        grid = view.make_model_grid()
        self.assertIn('key', grid.searchable_columns)
        self.assertIn('subject', grid.searchable_columns)

    def test_render_to_short(self):
        view = self.make_view()
        setting = EmailSetting(self.config)

        # more than 2 recips
        result = view.render_to_short(setting, 'to', [
            'alice@example.com',
            'bob@example.com',
            'charlie@example.com',
            'diana@example.com',
        ])
        self.assertEqual(result, 'alice@example.com, bob@example.com, ...')

        # just 2 recips
        result = view.render_to_short(setting, 'to', [
            'alice@example.com',
            'bob@example.com',
        ])
        self.assertEqual(result, 'alice@example.com, bob@example.com')

        # just 1 recip
        result = view.render_to_short(setting, 'to', ['alice@example.com'])
        self.assertEqual(result, 'alice@example.com')

        # no recips
        result = view.render_to_short(setting, 'to', [])
        self.assertIsNone(result)

    def test_get_instance(self):
        self.config.setdefault('wutta.email.default.sender', 'test@example.com')
        view = self.make_view()

        # normal
        with patch.object(self.request, 'matchdict', new={'key': 'feedback'}):
            setting = view.get_instance()
            self.assertIsInstance(setting, dict)
            self.assertIn('key', setting)
            self.assertIn('sender', setting)
            self.assertIn('subject', setting)
            self.assertIn('to', setting)
            self.assertIn('cc', setting)
            self.assertIn('notes', setting)
            self.assertIn('enabled', setting)

        # not found
        with patch.object(self.request, 'matchdict', new={'key': 'this-should_notEXIST'}):
            self.assertRaises(HTTPNotFound, view.get_instance)

    def test_get_instance_title(self):
        view = self.make_view()
        result = view.get_instance_title({'subject': 'whatever'})
        self.assertEqual(result, 'whatever')

    def test_configure_form(self):
        self.config.setdefault('wutta.email.default.sender', 'test@example.com')
        view = self.make_view()

        with patch.object(self.request, 'matchdict', new={'key': 'feedback'}):
            setting = view.get_instance()
            form = view.make_model_form(setting)
            self.assertIn('description', form.readonly_fields)
            self.assertFalse(form.required_fields['replyto'])

    def test_persist(self):
        model = self.app.model
        self.config.setdefault('wutta.email.default.sender', 'test@example.com')
        view = self.make_view()

        # start w/ no settings in db
        self.assertEqual(self.session.query(model.Setting).count(), 0)

        # "edit" settings for feedback email
        with patch.object(self.request, 'matchdict', new={'key': 'feedback'}):
            setting = view.get_instance()
            setting['subject'] = 'Testing Feedback'
            setting['sender'] = 'feedback@example.com'
            setting['replyto'] = 'feedback4@example.com'
            setting['to'] = 'feedback@example.com'
            setting['cc'] = 'feedback2@example.com'
            setting['bcc'] = 'feedback3@example.com'
            setting['notes'] = "did this work?"
            setting['enabled'] = True

            # persist email settings
            with patch.object(view, 'Session', return_value=self.session):
                view.persist(setting)
                self.session.commit()

        # check settings in db
        self.assertEqual(self.session.query(model.Setting).count(), 8)
        self.assertEqual(self.app.get_setting(self.session, 'wutta.email.feedback.subject'),
                         "Testing Feedback")
        self.assertEqual(self.app.get_setting(self.session, 'wutta.email.feedback.sender'),
                         'feedback@example.com')
        self.assertEqual(self.app.get_setting(self.session, 'wutta.email.feedback.replyto'),
                         'feedback4@example.com')
        self.assertEqual(self.app.get_setting(self.session, 'wutta.email.feedback.to'),
                         'feedback@example.com')
        self.assertEqual(self.app.get_setting(self.session, 'wutta.email.feedback.cc'),
                         'feedback2@example.com')
        self.assertEqual(self.app.get_setting(self.session, 'wutta.email.feedback.bcc'),
                         'feedback3@example.com')
        self.assertEqual(self.app.get_setting(self.session, 'wutta.email.feedback.notes'),
                         "did this work?")
        self.assertEqual(self.app.get_setting(self.session, 'wutta.email.feedback.enabled'),
                         'true')

        # "edit" settings for feedback email
        with patch.object(self.request, 'matchdict', new={'key': 'feedback'}):
            setting = view.get_instance()
            setting['subject'] = None
            setting['sender'] = None
            setting['replyto'] = None
            setting['to'] = None
            setting['cc'] = None
            setting['bcc'] = None
            setting['notes'] = None
            setting['enabled'] = False

            # persist email settings
            with patch.object(view, 'Session', return_value=self.session):
                view.persist(setting)
                self.session.commit()

        # check settings in db
        self.assertEqual(self.session.query(model.Setting).count(), 1)
        self.assertIsNone(self.app.get_setting(self.session, 'wutta.email.feedback.subject'))
        self.assertIsNone(self.app.get_setting(self.session, 'wutta.email.feedback.sender'))
        self.assertIsNone(self.app.get_setting(self.session, 'wutta.email.feedback.replyto'))
        self.assertIsNone(self.app.get_setting(self.session, 'wutta.email.feedback.to'))
        self.assertIsNone(self.app.get_setting(self.session, 'wutta.email.feedback.cc'))
        self.assertIsNone(self.app.get_setting(self.session, 'wutta.email.feedback.bcc'))
        self.assertIsNone(self.app.get_setting(self.session, 'wutta.email.feedback.notes'))
        self.assertEqual(self.app.get_setting(self.session, 'wutta.email.feedback.enabled'),
                         'false')

    def test_render_to_response(self):
        self.config.setdefault('wutta.email.default.sender', 'test@example.com')
        self.pyramid_config.add_route('home', '/')
        self.pyramid_config.add_route('login', '/auth/login')
        self.pyramid_config.add_route('email_settings', '/email/settings')
        self.pyramid_config.add_route('email_settings.preview', '/email/settings/{key}/preview')
        view = self.make_view()

        # nb. this gives coverage, but tests nothing..
        with patch.object(self.request, 'matchdict', new={'key': 'feedback'}):
            setting = view.get_instance()
            with patch.object(view, 'viewing', new=True):
                context = {'instance': setting}
                response = view.render_to_response('view', context)
                self.assertIsInstance(response, Response)

    def test_preview(self):
        self.config.setdefault('wutta.email.default.sender', 'test@example.com')
        view = self.make_view()

        # nb. this gives coverage, but tests nothing..
        with patch.object(self.request, 'matchdict', new={'key': 'feedback'}):

            # html
            with patch.object(self.request, 'params', new={'mode': 'html'}):
                response = view.preview()
                self.assertEqual(response.content_type, 'text/html')

            # txt
            with patch.object(self.request, 'params', new={'mode': 'txt'}):
                response = view.preview()
                self.assertEqual(response.content_type, 'text/plain')
